"""Identification of individual fragments given the predictions generate by the idCNN
"""
import logging

import numpy as np
from torch import nn

from idtrackerai import Fragment, ListOfFragments
from idtrackerai.network import NetworkParams

from .network.get_predictions import get_predictions_identities


def compute_identification_statistics_for_non_accumulated_fragments(
    fragments: list[Fragment],
    all_predictions: np.ndarray,
    all_softmax_probs: np.ndarray,
    number_of_animals=None,
):
    """Given the predictions associated to the images in each (individual)
    fragment in the list fragments if computes the statistics necessary for the
    identification of fragment.

    Parameters
    ----------
    fragments : list
        List of individual fragment objects
    assigner : <GetPrediction object>
        The assigner object has as main attributes the list of predictions
        associated to `images` and the the corresponding softmax vectors
    number_of_animals : int
        number of animals to be tracked
    """
    counter = 0
    for fragment in fragments:
        if not fragment.used_for_training and fragment.is_an_individual:
            next_counter_value = counter + fragment.number_of_images
            predictions = all_predictions[counter:next_counter_value]
            softmax_probs = all_softmax_probs[counter:next_counter_value]
            fragment.compute_identification_statistics(
                predictions, softmax_probs, number_of_animals=number_of_animals
            )
            counter = next_counter_value


def assign_identity(list_of_fragments: ListOfFragments):
    """Identifies the individual fragments recursively, based on the value of
    P2

    Parameters
    ----------
    list_of_fragments : <ListOfFragments object>
        collection of the individual fragments and associated methods
    """
    logging.info("Assigning identities")
    list_of_fragments.compute_P2_vectors()
    fragment = list_of_fragments.get_next_fragment_to_identify()
    while fragment:
        fragment.assign_identity()
        fragment = list_of_fragments.get_next_fragment_to_identify()


def assign_remaining_fragments(
    list_of_fragments: ListOfFragments,
    identification_model: nn.Module,
    network_params: NetworkParams,
):
    """This is the main function of this module: given a list_of_fragments it
    puts in place the routine to identify, if possible, each of the individual
    fragments. The starting point for the identification is given by the
    predictions produced by the ConvNetwork net passed as input. The organisation
    of the images in individual fragments is then used to assign more accurately.

    Parameters
    ----------
    list_of_fragments : <ListOfFragments object>
        collection of the individual fragments and associated methods
    video : <Video object>
        Object collecting all the parameters of the video and paths for saving and loading
    net : <ConvNetwork object>
        Convolutional neural network object created according to net.params

    See Also
    --------
    ListOfFragments.get_images_from_fragments_to_assign
    assign
    compute_identification_statistics_for_non_accumulated_fragments

    """
    logging.info("Assigning identities to all non-accumulated individual fragments")
    list_of_fragments.reset(roll_back_to="accumulation")
    number_of_unidentified_individual_fragments = (
        list_of_fragments.get_number_of_unidentified_individual_fragments()
    )
    logging.info(
        "Number of unidentified individual fragments: "
        f"{number_of_unidentified_individual_fragments}"
    )
    if not number_of_unidentified_individual_fragments:
        list_of_fragments.compute_P2_vectors()
        return

    images = list_of_fragments.get_images_from_fragments_to_assign()

    predictions, softmax_probs = get_predictions_identities(
        identification_model, images, network_params
    )

    logging.debug(
        f"{len(predictions)} generated predictions between "
        f"identities {set(predictions)}"
    )
    compute_identification_statistics_for_non_accumulated_fragments(
        list_of_fragments.fragments, predictions, softmax_probs
    )
    assign_identity(list_of_fragments)



import logging
import shutil
import sys
from importlib.metadata import version
from importlib.resources import files
from pathlib import Path

import toml
from PyQt6.QtWidgets import QApplication

from idtrackerai.utils import (
    check_version_on_console_thread,
    conf,
    initLogger,
    pprint_dict,
)

from .arg_parser import parse_args

# PyQt has to be imported before CV2 (importing idtrackerai stuff implies CV2)
# If not, the QFileDialog.getFileNames() does not load the icons, very weird

all_valid_parameters = (
    (Path(__file__).parent / "all_valid_parameters.dat").read_text().splitlines()
)


def load_toml(path: Path, name: str = "") -> dict:
    if not path.is_file():
        raise FileNotFoundError(f"{path} do not exist")
    try:
        toml_dict = {
            key.lower(): value for key, value in toml.load(path.open()).items()
        }

        invalid_keys = [
            key for key in toml_dict.keys() if key not in all_valid_parameters
        ]

        if invalid_keys:
            logging.error(
                f"Not recognized parameters while reading {path}: {invalid_keys}"
            )
            sys.exit()

        for key, value in toml_dict.items():
            if value == "":
                toml_dict[key] = None
        if name:
            logging.info(pprint_dict(toml_dict, name), extra={"markup": True})
        return toml_dict
    except Exception:
        logging.error(f"Could not read {path}, bad format")
        sys.exit()


def main() -> bool:
    """The command `idtrackerai` runs this function"""
    parameters = {}
    initLogger(check_version=False)

    constants = load_toml((files("idtrackerai") / "constants.toml"))  # type: ignore
    parameters.update(constants)

    if Path("local_settings.py").is_file():
        logging.warning("Deprecated local_settings format found in ./local_settings.py")

    local_settings_path = Path("local_settings.toml")
    if local_settings_path.is_file():
        local_settings_dict = load_toml(local_settings_path, "Local settings")
        parameters.update(local_settings_dict)

    conf.set_dict(constants)  # TODO for weird parameters for GUI
    terminal_args = parse_args(constants)
    ready_to_track = terminal_args.pop("track")

    if "general_settings" in terminal_args:
        general_settings = load_toml(
            terminal_args.pop("general_settings"), "General settings"
        )
        parameters.update(general_settings)
    else:
        logging.info("No general settings loaded")

    if "session_parameters" in terminal_args:
        session_parameters = load_toml(
            terminal_args.pop("session_parameters"), "Session parameters"
        )
        parameters.update(session_parameters)
    else:
        logging.info("No session parameters loaded")

    if terminal_args:
        logging.info(
            pprint_dict(terminal_args, "Terminal arguments"), extra={"markup": True}
        )
        parameters.update(terminal_args)
    else:
        logging.info("No terminal arguments detected")

    if ready_to_track:
        from .run_idtrackerai import RunIdTrackerAi

        check_version_on_console_thread()

        return RunIdTrackerAi(parameters).track_video()
    run_segmentation_GUI(parameters)
    if parameters.get("run_idtrackerai"):
        from .run_idtrackerai import RunIdTrackerAi

        return RunIdTrackerAi(parameters).track_video()
    return False


def run_segmentation_GUI(params: dict):
    from idtrackerai_start_app.segmentation_GUI import SegmentationGUI

    app = QApplication(sys.argv)
    window = SegmentationGUI(params)
    window.show()
    app.exec()


def general_test():
    from time import perf_counter

    from .run_idtrackerai import RunIdTrackerAi

    COMPRESSED_VIDEO_PATH = Path(str(files("idtrackerai"))) / "data" / "test_B.avi"

    video_path = Path.cwd() / COMPRESSED_VIDEO_PATH.name
    shutil.copyfile(COMPRESSED_VIDEO_PATH, video_path)

    initLogger(testing=True)

    params = load_toml((files("idtrackerai") / "constants.toml"))  # type: ignore
    params.update(
        {
            "session": "test",
            "video_paths": video_path,
            "tracking_intervals": None,
            "intensity_ths": [0, 155],
            "area_ths": [150, 60000],
            "number_of_animals": 8,
            "resolution_reduction": 1.0,
            "check_segmentation": False,
            "ROI_list": None,
            "track_wo_identities": False,
            "use_bkg": False,
            "protocol3_action": "continue",
        }
    )

    start = perf_counter()
    success = RunIdTrackerAi(params).track_video()
    minutes = (perf_counter() - start) / 60
    if success:
        logging.info(
            "[green]Test passed successfully in %.2f min with version %s",
            minutes,
            version("idtrackerai"),
            extra={"markup": True},
        )


if __name__ == "__main__":
    main()



# This file is part of idtracker.ai a multiple animals tracking system
# described in [1].
# Copyright (C) 2017- Francisco Romero Ferrero, Mattia G. Bergomi,
# Francisco J.H. Heras, Robert Hinz, Gonzalo G. de Polavieja and the
# Champalimaud Foundation.
#
# idtracker.ai is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details. In addition, we require
# derivatives or applications to acknowledge the authors by citing [1].
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# For more information please send an email (idtrackerai@gmail.com) or
# use the tools available at https://gitlab.com/polavieja_lab/idtrackerai.git.
#
# [1] Romero-Ferrero, F., Bergomi, M.G., Hinz, R.C., Heras, F.J.H.,
# de Polavieja, G.G., Nature Methods, 2019.
# idtracker.ai: tracking all individuals in small or large collectives of
# unmarked animals.
# (F.R.-F. and M.G.B. contributed equally to this work.
# Correspondence should be addressed to G.G.d.P:
# gonzalo.polavieja@neuro.fchampalimaud.org)
import logging

import torch
from torch import nn
from torch.backends import cudnn
from torch.optim.lr_scheduler import MultiStepLR

from idtrackerai import Video
from idtrackerai.network import LearnerClassification, NetworkParams
from idtrackerai.utils import conf

from .accumulation_manager import (
    AccumulationManager,
    get_predictions_of_candidates_fragments,
)
from .dataset.identification_dataloader import get_training_data_loaders
from .dataset.identification_dataset import split_data_train_and_validation
from .network.stop_training_criteria import StopTraining
from .network.trainer import TrainIdentification


def perform_one_accumulation_step(
    accumulation_manager: AccumulationManager,
    video: Video,
    identification_model: nn.Module,
    learner_class: type[LearnerClassification],
    network_params: NetworkParams,
):
    logging.info(
        f"[bold]Performing new accumulation, step {accumulation_manager.counter}",
        extra={"markup": True},
    )
    video.accumulation_step = accumulation_manager.counter

    # Get images for training
    accumulation_manager.get_new_images_and_labels()
    images, labels = accumulation_manager.get_images_and_labels_for_training()
    train_data, val_data = split_data_train_and_validation(
        images, labels, validation_proportion=conf.VALIDATION_PROPORTION
    )
    assert images.shape[0] == labels.shape[0]
    logging.info(
        f"Training with {len(train_data['images'])}, "
        f"validating with {len(val_data['images'])}"
    )
    assert len(val_data["images"]) > 0

    # Set data loaders
    train_loader, val_loader = get_training_data_loaders(
        video.number_of_animals, train_data, val_data
    )

    # Set criterion
    logging.info("Setting training criterion")
    criterion = nn.CrossEntropyLoss(weight=torch.tensor(train_data["weights"]))

    # Send model and criterion to GPU
    if network_params.use_gpu:
        torch.cuda.set_device(0)
        logging.info(
            'Sending model and criterion to GPU: "%s"', torch.cuda.get_device_name()
        )
        cudnn.benchmark = True  # make it train faster
        identification_model = identification_model.cuda()
        criterion = criterion.cuda()

    # Set optimizer
    logging.info("Setting optimizer")
    optimizer = torch.optim.__dict__[network_params.optimizer](
        identification_model.parameters(), **network_params.optim_args
    )

    # Set scheduler
    logging.info("Setting scheduler")
    scheduler = MultiStepLR(optimizer, milestones=network_params.schedule, gamma=0.1)

    # Set learner
    logging.info("Setting the learner")
    learner = learner_class(identification_model, criterion, optimizer, scheduler)

    # Set stopping criteria
    logging.info("Setting the stopping criteria")
    # set criteria to stop the training
    stop_training = StopTraining(
        network_params.number_of_classes,
        check_for_loss_plateau=True,
        first_accumulation_flag=video is None or video.accumulation_step == 0,
    )

    TrainIdentification(
        learner,
        train_loader,
        val_loader,
        network_params,
        stop_training,
        accumulation_manager=accumulation_manager,
    )
    logging.info("Identification network trained")

    accumulation_manager.update_used_images_and_labels()
    accumulation_manager.assign_identities_to_fragments_used_for_training()
    accumulation_manager.update_list_of_individual_fragments_used()

    # compute ratio of accumulated images and stop if it is above random
    accumulation_manager.ratio_accumulated_images = (
        accumulation_manager.list_of_fragments.compute_ratio_of_images_used_for_training()
    )

    if (
        accumulation_manager.ratio_accumulated_images
        > conf.THRESHOLD_EARLY_STOP_ACCUMULATION
    ):
        logging.debug("Stopping accumulation by early stopping criteria")
        return accumulation_manager.ratio_accumulated_images

    # Set accumulation parameters for rest of the accumulation
    # take images from global fragments not used in training (in the remainder test global fragments)
    if any(
        not global_fragment.used_for_training
        for global_fragment in accumulation_manager.list_of_global_fragments.global_fragments
    ):
        logging.info(
            "Generating [bold]predictions[/bold] on remaining global fragments",
            extra={"markup": True},
        )
        (
            predictions,
            softmax_probs,
            indices_to_split,
            candidate_individual_fragments_identifiers,
        ) = get_predictions_of_candidates_fragments(
            identification_model,
            video.id_images_file_paths,
            network_params,
            accumulation_manager.list_of_fragments.fragments,
        )

        accumulation_manager.split_predictions_after_network_assignment(
            predictions,
            softmax_probs,
            indices_to_split,
            candidate_individual_fragments_identifiers,
        )
        # assign identities to the global fragments based on the predictions
        logging.info(
            "Checking eligibility criteria and generate the "
            "new list of identified global fragments to accumulate"
        )
        accumulation_manager.get_acceptable_global_fragments_for_training(
            candidate_individual_fragments_identifiers, video.accumulation_trial
        )

        accumulation_manager.print_accumulation_variables()

        stats = video.accumulation_statistics

        stats["n_accumulated_global_fragments"].append(
            sum(
                global_fragment.used_for_training
                for global_fragment in accumulation_manager.list_of_global_fragments.global_fragments
            )
        )
        stats["n_non_certain_global_fragments"].append(
            accumulation_manager.number_of_noncertain_global_fragments
        )
        stats["n_randomly_assigned_global_fragments"].append(
            accumulation_manager.number_of_random_assigned_global_fragments
        )
        stats["n_nonconsistent_global_fragments"].append(
            accumulation_manager.number_of_nonconsistent_global_fragments
        )
        stats["n_nonunique_global_fragments"].append(
            accumulation_manager.number_of_nonunique_global_fragments
        )
        stats["n_acceptable_global_fragments"].append(
            sum(
                global_fragment.acceptable_for_training(
                    accumulation_manager.accumulation_strategy
                )
                for global_fragment in accumulation_manager.list_of_global_fragments.global_fragments
            )
        )
        stats["ratio_of_accumulated_images"].append(
            accumulation_manager.ratio_accumulated_images
        )

        accumulation_manager.update_counter()

    accumulation_manager.ratio_accumulated_images = (
        accumulation_manager.list_of_fragments.compute_ratio_of_images_used_for_training()
    )

    video.accumulation_statistics_data[video.accumulation_trial] = (
        video.accumulation_statistics
    )

    return accumulation_manager.ratio_accumulated_images



from PyQt6.QtGui import QPalette

ColorRole = QPalette.ColorRole
ColorGroup = QPalette.ColorGroup
Disabled = QPalette.ColorGroup.Disabled

custom = QPalette()
background = 0x202124
mid_background = 0x2C2D2F
light_bkg = 0x3F4042
blue = 0x8AB4F7
blue = 0xAAD4FF
almost_white = 0xFDFDFD
placeholder_color = 0xB0B0B0
red = 0xFF0000
custom.setColor(ColorRole.Window, background)
custom.setColor(ColorRole.WindowText, almost_white)
custom.setColor(ColorRole.Base, light_bkg)
custom.setColor(ColorRole.AlternateBase, mid_background)
custom.setColor(ColorRole.Text, almost_white)
custom.setColor(ColorRole.Button, mid_background)
custom.setColor(ColorRole.ButtonText, blue)
custom.setColor(ColorRole.Highlight, blue)
custom.setColor(ColorRole.HighlightedText, background)
custom.setColor(ColorRole.PlaceholderText, placeholder_color)
custom.setColor(Disabled, ColorRole.WindowText, custom.windowText().color().darker())
custom.setColor(Disabled, ColorRole.Button, custom.button().color().darker())
custom.setColor(Disabled, ColorRole.ButtonText, custom.buttonText().color().darker())
custom.setColor(Disabled, ColorRole.Base, custom.base().color().darker())
custom.setColor(Disabled, ColorRole.Text, custom.text().color().darker())
custom.setColor(
    Disabled, ColorRole.PlaceholderText, custom.placeholderText().color().darker()
)
custom.setColor(ColorRole.ToolTipBase, 0xFFFFDC)
custom.setColor(ColorRole.ToolTipText, 0x000000)
custom.setColor(ColorRole.Dark, placeholder_color)
custom.setColor(ColorRole.Light, mid_background)
custom.setColor(ColorRole.BrightText, red)
custom.setColor(ColorRole.Midlight, red)
custom.setColor(ColorRole.Shadow, red)
custom.setColor(ColorRole.Mid, red)
custom.setColor(ColorRole.Link, red)
custom.setColor(ColorRole.LinkVisited, red)
custom.setColor(ColorRole.NoRole, red)


dark = QPalette()
dark.setColor(ColorGroup.Active, ColorRole.WindowText, 0xF0F0F0)
dark.setColor(ColorGroup.Active, ColorRole.Button, 0x323232)
dark.setColor(ColorGroup.Active, ColorRole.Light, 0x4B4B4B)
dark.setColor(ColorGroup.Active, ColorRole.Midlight, 0x2A2A2A)
dark.setColor(ColorGroup.Active, ColorRole.Dark, 0x212121)
dark.setColor(ColorGroup.Active, ColorRole.Mid, 0x262626)
dark.setColor(ColorGroup.Active, ColorRole.Text, 0xF0F0F0)
dark.setColor(ColorGroup.Active, ColorRole.BrightText, 0x4B4B4B)
dark.setColor(ColorGroup.Active, ColorRole.ButtonText, blue)
dark.setColor(ColorGroup.Active, ColorRole.Base, 0x242424)
dark.setColor(ColorGroup.Active, ColorRole.Window, 0x323232)
dark.setColor(ColorGroup.Active, ColorRole.Shadow, 0x191919)
dark.setColor(ColorGroup.Active, ColorRole.Highlight, 0x308CC6)
dark.setColor(ColorGroup.Active, ColorRole.HighlightedText, 0xF0F0F0)
dark.setColor(ColorGroup.Active, ColorRole.Link, 0x308CC6)
dark.setColor(ColorGroup.Active, ColorRole.LinkVisited, 0xFF00FF)
dark.setColor(ColorGroup.Active, ColorRole.AlternateBase, 0x2B2B2B)
dark.setColor(ColorGroup.Active, ColorRole.ToolTipBase, 0xFFFFDC)
dark.setColor(ColorGroup.Active, ColorRole.ToolTipText, 0x000000)
dark.setColor(ColorGroup.Active, ColorRole.PlaceholderText, 0xA0A0A0)
dark.setColor(ColorGroup.Active, ColorRole.NoRole, 0x000000)
dark.setColor(ColorGroup.Disabled, ColorRole.WindowText, 0x828282)
dark.setColor(ColorGroup.Disabled, ColorRole.Button, 0x323232)
dark.setColor(ColorGroup.Disabled, ColorRole.Light, 0x4B4B4B)
dark.setColor(ColorGroup.Disabled, ColorRole.Midlight, 0x2A2A2A)
dark.setColor(ColorGroup.Disabled, ColorRole.Dark, 0xBEBEBE)
dark.setColor(ColorGroup.Disabled, ColorRole.Mid, 0x262626)
dark.setColor(ColorGroup.Disabled, ColorRole.Text, 0x828282)
dark.setColor(ColorGroup.Disabled, ColorRole.BrightText, 0x4B4B4B)
dark.setColor(ColorGroup.Disabled, ColorRole.ButtonText, 0x828282)
dark.setColor(ColorGroup.Disabled, ColorRole.Base, 0x323232)
dark.setColor(ColorGroup.Disabled, ColorRole.Window, 0x323232)
dark.setColor(ColorGroup.Disabled, ColorRole.Shadow, 0x252525)
dark.setColor(ColorGroup.Disabled, ColorRole.Highlight, 0x919191)
dark.setColor(ColorGroup.Disabled, ColorRole.HighlightedText, 0xF0F0F0)
dark.setColor(ColorGroup.Disabled, ColorRole.Link, 0x308CC6)
dark.setColor(ColorGroup.Disabled, ColorRole.LinkVisited, 0xFF00FF)
dark.setColor(ColorGroup.Disabled, ColorRole.AlternateBase, 0x2B2B2B)
dark.setColor(ColorGroup.Disabled, ColorRole.ToolTipBase, 0xFFFFDC)
dark.setColor(ColorGroup.Disabled, ColorRole.ToolTipText, 0x000000)
dark.setColor(ColorGroup.Disabled, ColorRole.PlaceholderText, 0x4B4B4B)
dark.setColor(ColorGroup.Disabled, ColorRole.NoRole, 0x000000)
dark.setColor(ColorGroup.Inactive, ColorRole.WindowText, 0xF0F0F0)
dark.setColor(ColorGroup.Inactive, ColorRole.Button, 0x323232)
dark.setColor(ColorGroup.Inactive, ColorRole.Light, 0x4B4B4B)
dark.setColor(ColorGroup.Inactive, ColorRole.Midlight, 0x2A2A2A)
dark.setColor(ColorGroup.Inactive, ColorRole.Dark, 0x212121)
dark.setColor(ColorGroup.Inactive, ColorRole.Mid, 0x262626)
dark.setColor(ColorGroup.Inactive, ColorRole.Text, 0xF0F0F0)
dark.setColor(ColorGroup.Inactive, ColorRole.BrightText, 0x4B4B4B)
dark.setColor(ColorGroup.Inactive, ColorRole.ButtonText, 0xF0F0F0)
dark.setColor(ColorGroup.Inactive, ColorRole.Base, 0x242424)
dark.setColor(ColorGroup.Inactive, ColorRole.Window, 0x323232)
dark.setColor(ColorGroup.Inactive, ColorRole.Shadow, 0x191919)
dark.setColor(ColorGroup.Inactive, ColorRole.Highlight, 0x308CC6)
dark.setColor(ColorGroup.Inactive, ColorRole.HighlightedText, 0xF0F0F0)
dark.setColor(ColorGroup.Inactive, ColorRole.Link, 0x308CC6)
dark.setColor(ColorGroup.Inactive, ColorRole.LinkVisited, 0xFF00FF)
dark.setColor(ColorGroup.Inactive, ColorRole.AlternateBase, 0x2B2B2B)
dark.setColor(ColorGroup.Inactive, ColorRole.ToolTipBase, 0xFFFFDC)
dark.setColor(ColorGroup.Inactive, ColorRole.ToolTipText, 0x000000)
dark.setColor(ColorGroup.Inactive, ColorRole.PlaceholderText, 0xF0F0F0)
dark.setColor(ColorGroup.Inactive, ColorRole.NoRole, 0x000000)


light = QPalette()
light.setColor(ColorGroup.Active, ColorRole.WindowText, 0x000000)
light.setColor(ColorGroup.Active, ColorRole.Button, 0xEFEFEF)
light.setColor(ColorGroup.Active, ColorRole.Light, 0xFFFFFF)
light.setColor(ColorGroup.Active, ColorRole.Midlight, 0xCACACA)
light.setColor(ColorGroup.Active, ColorRole.Dark, 0x9F9F9F)
light.setColor(ColorGroup.Active, ColorRole.Mid, 0xB8B8B8)
light.setColor(ColorGroup.Active, ColorRole.Text, 0x000000)
light.setColor(ColorGroup.Active, ColorRole.BrightText, 0xFFFFFF)
light.setColor(ColorGroup.Active, ColorRole.ButtonText, 0x000000)
light.setColor(ColorGroup.Active, ColorRole.Base, 0xFFFFFF)
light.setColor(ColorGroup.Active, ColorRole.Window, 0xEFEFEF)
light.setColor(ColorGroup.Active, ColorRole.Shadow, 0x767676)
light.setColor(ColorGroup.Active, ColorRole.Highlight, 0x308CC6)
light.setColor(ColorGroup.Active, ColorRole.HighlightedText, 0xFFFFFF)
light.setColor(ColorGroup.Active, ColorRole.Link, 0x0000FF)
light.setColor(ColorGroup.Active, ColorRole.LinkVisited, 0xFF00FF)
light.setColor(ColorGroup.Active, ColorRole.AlternateBase, 0xF7F7F7)
light.setColor(ColorGroup.Active, ColorRole.ToolTipBase, 0xFFFFDC)
light.setColor(ColorGroup.Active, ColorRole.ToolTipText, 0x000000)
light.setColor(ColorGroup.Active, ColorRole.PlaceholderText, 0x404040)
light.setColor(ColorGroup.Active, ColorRole.NoRole, 0x000000)
light.setColor(ColorGroup.Disabled, ColorRole.WindowText, 0xBEBEBE)
light.setColor(ColorGroup.Disabled, ColorRole.Button, 0xEFEFEF)
light.setColor(ColorGroup.Disabled, ColorRole.Light, 0xFFFFFF)
light.setColor(ColorGroup.Disabled, ColorRole.Midlight, 0xCACACA)
light.setColor(ColorGroup.Disabled, ColorRole.Dark, 0xBEBEBE)
light.setColor(ColorGroup.Disabled, ColorRole.Mid, 0xB8B8B8)
light.setColor(ColorGroup.Disabled, ColorRole.Text, 0xBEBEBE)
light.setColor(ColorGroup.Disabled, ColorRole.BrightText, 0xFFFFFF)
light.setColor(ColorGroup.Disabled, ColorRole.ButtonText, 0xBEBEBE)
light.setColor(ColorGroup.Disabled, ColorRole.Base, 0xEFEFEF)
light.setColor(ColorGroup.Disabled, ColorRole.Window, 0xEFEFEF)
light.setColor(ColorGroup.Disabled, ColorRole.Shadow, 0xB1B1B1)
light.setColor(ColorGroup.Disabled, ColorRole.Highlight, 0x919191)
light.setColor(ColorGroup.Disabled, ColorRole.HighlightedText, 0xFFFFFF)
light.setColor(ColorGroup.Disabled, ColorRole.Link, 0x0000FF)
light.setColor(ColorGroup.Disabled, ColorRole.LinkVisited, 0xFF00FF)
light.setColor(ColorGroup.Disabled, ColorRole.AlternateBase, 0xF7F7F7)
light.setColor(ColorGroup.Disabled, ColorRole.ToolTipBase, 0xFFFFDC)
light.setColor(ColorGroup.Disabled, ColorRole.ToolTipText, 0x000000)
light.setColor(ColorGroup.Disabled, ColorRole.PlaceholderText, 0xC0C0C0)
light.setColor(ColorGroup.Disabled, ColorRole.NoRole, 0x000000)
light.setColor(ColorGroup.Inactive, ColorRole.WindowText, 0x000000)
light.setColor(ColorGroup.Inactive, ColorRole.Button, 0xEFEFEF)
light.setColor(ColorGroup.Inactive, ColorRole.Light, 0xFFFFFF)
light.setColor(ColorGroup.Inactive, ColorRole.Midlight, 0xCACACA)
light.setColor(ColorGroup.Inactive, ColorRole.Dark, 0x9F9F9F)
light.setColor(ColorGroup.Inactive, ColorRole.Mid, 0xB8B8B8)
light.setColor(ColorGroup.Inactive, ColorRole.Text, 0x000000)
light.setColor(ColorGroup.Inactive, ColorRole.BrightText, 0xFFFFFF)
light.setColor(ColorGroup.Inactive, ColorRole.ButtonText, 0x000000)
light.setColor(ColorGroup.Inactive, ColorRole.Base, 0xFFFFFF)
light.setColor(ColorGroup.Inactive, ColorRole.Window, 0xEFEFEF)
light.setColor(ColorGroup.Inactive, ColorRole.Shadow, 0x767676)
light.setColor(ColorGroup.Inactive, ColorRole.Highlight, 0x308CC6)
light.setColor(ColorGroup.Inactive, ColorRole.HighlightedText, 0xFFFFFF)
light.setColor(ColorGroup.Inactive, ColorRole.Link, 0x0000FF)
light.setColor(ColorGroup.Inactive, ColorRole.LinkVisited, 0xFF00FF)
light.setColor(ColorGroup.Inactive, ColorRole.AlternateBase, 0xF7F7F7)
light.setColor(ColorGroup.Inactive, ColorRole.ToolTipBase, 0xFFFFDC)
light.setColor(ColorGroup.Inactive, ColorRole.ToolTipText, 0x000000)
light.setColor(ColorGroup.Inactive, ColorRole.PlaceholderText, 0x000000)
light.setColor(ColorGroup.Inactive, ColorRole.NoRole, 0x000000)



import json
import logging
from dataclasses import asdict, dataclass, field
from pathlib import Path
from typing import Optional

from idtrackerai.utils import create_dir, json_default


@dataclass
class NetworkParams:
    number_of_classes: int
    schedule: list[int]
    architecture: str
    model_name: str
    dataset: str
    image_size: list[int]
    optim_args: dict = field(default_factory=dict)
    epochs: int = 0
    optimizer: str = "SGD"
    loss: str = "CE"
    use_gpu: bool = True
    apply_mask: bool = False
    skip_eval: bool = False
    save_folder: Path = Path("")

    scopes_layers_to_optimize: Optional[list[str]] = field(default_factory=list)
    knowledge_transfer_folder: Path | None = None
    use_adam_optimiser: bool = False
    return_store_objects: bool = False
    restore_folder: Path = Path()

    @property
    def load_model_path(self) -> Path:
        # v5.0.0 compatibility
        v5_path = self.restore_folder / (self.model_file_name + "_.model.pth")
        if v5_path.is_file():
            return v5_path
        return self.restore_folder / (self.model_file_name + ".model.pth")

    @property
    def save_model_path(self) -> Path:
        return self.save_folder / self.model_file_name

    @property
    def model_file_name(self) -> str:
        return f"{self.dataset}_{self.model_name}"

    @property
    def knowledge_transfer_model_file(self) -> Path | None:
        if self.knowledge_transfer_folder is None:
            return None
        return (
            self.knowledge_transfer_folder
            / "supervised_identification_network.model.pth"
        )

    def save(self) -> None:
        path = self.save_folder / "model_params.json"
        logging.info(f"Saving NetworkParams at {path}")
        create_dir(self.save_folder)
        path.write_text(json.dumps(asdict(self), indent=4, default=json_default))



# This file is part of idtracker.ai a multiple animals tracking system
# described in [1].
# Copyright (C) 2017- Francisco Romero Ferrero, Mattia G. Bergomi,
# Francisco J.H. Heras, Robert Hinz, Gonzalo G. de Polavieja and the
# Champalimaud Foundation.
#
# idtracker.ai is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details. In addition, we require
# derivatives or applications to acknowledge the authors by citing [1].
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# For more information please send an email (idtrackerai@gmail.com) or
# use the tools available at https://gitlab.com/polavieja_lab/idtrackerai.git.
#
# [1] Romero-Ferrero, F., Bergomi, M.G., Hinz, R.C., Heras, F.J.H.,
# de Polavieja, G.G., Nature Methods, 2019.
# idtracker.ai: tracking all individuals in small or large collectives of
# unmarked animals.
# (F.R.-F. and M.G.B. contributed equally to this work.
# Correspondence should be addressed to G.G.d.P:
# gonzalo.polavieja@neuro.fchampalimaud.org)
import json
import logging
from dataclasses import dataclass
from pathlib import Path
from shutil import rmtree
from time import perf_counter
from typing import Iterable, Optional, TypeVar

import cv2
import h5py
import numpy as np
from rich.progress import BarColumn, Progress, TaskProgressColumn, TimeRemainingColumn

InputType = TypeVar("InputType")


def track(
    sequence: Iterable[InputType],  # TODO also Sequence?
    desc: str = "Working...",
    total: Optional[float] = None,
) -> Iterable[InputType]:
    """A custom interpretation of rich.progress.track"""

    progress = Progress(
        " " * 18 + desc,
        BarColumn(bar_width=None),
        TaskProgressColumn(show_speed=True),
        TimeRemainingColumn(elapsed_when_finished=True),
    )

    with progress:
        yield from progress.track(sequence, total, description=desc)


def delete_attributes_from_object(object_to_modify, list_of_attributes):
    for attribute in list_of_attributes:
        if hasattr(object_to_modify, attribute):
            delattr(object_to_modify, attribute)


def create_dir(path: Path, remove_existing=False):
    if path.is_dir():
        if remove_existing:
            rmtree(path)
            path.mkdir()
            logging.info(f"Directory {path} has been cleaned")
        else:
            logging.info(f"Directory {path} already exists")
    else:
        if not path.parent.is_dir():
            path.parent.mkdir()
        path.mkdir()
        logging.info(f"Directory {path} has been created")


def remove_dir(path: Path):
    if path.is_dir():
        rmtree(path, ignore_errors=True)
        logging.info(f"Directory {path} has been removed")
    else:
        logging.info(f"Directory {path} not found, can't remove")


def remove_file(path: Path):
    if path.is_file():
        path.unlink()
        logging.info(f"File {path} has been removed")


def assert_all_files_exist(paths: list[Path]):
    """Returns FileNotFoundError if any of the paths is not an existing file"""
    for path in paths:
        if not path.is_file():
            raise FileNotFoundError(f"File {path} not found")


def get_vertices_from_label(label: str, close=False):
    """Transforms a string representation of a polygon from the
    ROI widget (idtrackerai_app) into a vertices np.array"""
    data = json.loads(label[10:].replace("'", '"'))

    if label[2:9] == "Polygon":
        vertices = np.asarray(data)
    elif label[2:9] == "Ellipse":
        vertices = cv2.ellipse2Poly(
            data["center"], data["axes"], data["angle"], 0, 360, 2
        )
    else:
        raise TypeError(label)

    if close:
        return np.vstack([vertices, vertices[0]]).astype(np.int32)
    return vertices.astype(np.int32)


def build_ROI_mask_from_list(
    list_of_ROIs: None | list[str] | str, width, height
) -> np.ndarray | None:
    """Transforms a list of polygons (as type str) from
    ROI widget (idtrackerai_app) into a boolean np.array mask"""

    if list_of_ROIs is None:
        return None

    ROI_mask = np.zeros((height, width), np.uint8)

    if isinstance(list_of_ROIs, str):
        list_of_ROIs = list(list_of_ROIs)

    for line in list_of_ROIs:
        vertices = get_vertices_from_label(line)
        if line[0] == "+":
            cv2.fillPoly(ROI_mask, [vertices][::-1], color=1)
        elif line[0] == "-":
            cv2.fillPoly(ROI_mask, [vertices][::-1], color=0)
        else:
            raise TypeError
    return ROI_mask.astype(bool)


class CustomError(Exception):
    pass


@dataclass
class Episode:
    index: int
    local_start: int
    local_end: int
    video_path_index: int
    global_start: int
    global_end: int


class Timer:
    """Simple class for measuring execution time during the whole process"""

    name: str
    interval: float = -1.0
    start_time: float = -1.0

    def __init__(self, name: str = ""):
        self.name = name
        self.reset()

    def reset(self):
        self.interval = -1
        self.start_time = -1

    @property
    def started(self):
        return self.start_time > 0

    @property
    def finished(self):
        return self.interval > 0

    def start(self):
        logging.info("[blue bold]START %s", self.name, extra={"markup": True})
        self.start_time = perf_counter()

    def finish(self, raise_if_not_started=True) -> float:
        if not self.started:
            if raise_if_not_started:
                raise RuntimeError("Timer finish method called before start method")
            return -1

        self.interval = perf_counter() - self.start_time

        logging.info(
            f"[blue bold]FINISH {self.name}, it took {self}", extra={"markup": True}
        )
        return self.interval

    def __str__(self) -> str:
        if self.interval > 6000:
            return f"{self.interval/3600:.4f} hours"
        if self.interval > 100:
            return f"{self.interval/60:.4f} minutes"
        if self.interval > 0:
            return f"{self.interval:.4f} seconds"
        if self.started:
            return "not finished"
        return "not started"

    @classmethod
    def from_dict(cls, d: dict):
        obj = cls.__new__(cls)
        obj.__dict__.update(d)
        return obj


def check_if_identity_transfer_is_possible(
    number_of_animals: int, knowledge_transfer_folder: Path | None
) -> tuple[bool, list[int]]:
    if knowledge_transfer_folder is None:
        raise CustomError(
            "To perform identity transfer you "
            "need to provide a path for the variable "
            "'KNOWLEDGE_TRANSFER_FOLDER'"
        )

    kt_info_dict_path = knowledge_transfer_folder / "model_params.json"
    if kt_info_dict_path.is_file():
        knowledge_transfer_info_dict = json.load(kt_info_dict_path.open())
        assert "image_size" in knowledge_transfer_info_dict

    elif kt_info_dict_path.with_suffix(".npy").is_file():
        knowledge_transfer_info_dict: dict = np.load(
            kt_info_dict_path.with_suffix(".npy"), allow_pickle=True
        ).item()  # loading from v4
        assert "image_size" in knowledge_transfer_info_dict
    else:
        raise CustomError(
            "To perform identity transfer the models_params.npy file "
            "is needed to check the input_image_size and "
            "the number_of_classes of the model to be loaded"
        )
    is_identity_transfer_possible = (
        number_of_animals == knowledge_transfer_info_dict["number_of_classes"]
    )
    if is_identity_transfer_possible:
        logging.info(
            "Tracking with identity transfer. "
            "The identification_image_size will be matched "
            "to the image_size of the transferred network"
        )
        id_image_size = knowledge_transfer_info_dict["image_size"]
    else:
        logging.warning(
            "Tracking with identity transfer is not possible. "
            "The number of animals in the video needs to be the same as "
            "the number of animals in the transferred network"
        )
        id_image_size = []

    return is_identity_transfer_possible, id_image_size


def pprint_dict(d: dict, name: str = "") -> str:
    text = f"[bold blue]{name}[/]:" if name else ""

    pad = min(max(map(len, d.keys())), 25)

    for key, value in d.items():
        if isinstance(value, tuple):
            value = list(value)
        if isinstance(value, list) and value and isinstance(value[0], Path):
            value = list(map(str, value))
        if isinstance(value, Path):
            value = str(value)
        if len(repr(value)) < 50 or not isinstance(value, list):
            text += f"\n[bold]{key:>{pad}}[/] = {repr(value)}"
        else:
            s = f"[{repr(value[0])}"
            for item in value[1:]:
                s += f",\n{' '*pad}    {repr(item)}"
            s += "]"
            text += f"\n[bold]{key:>{pad}}[/] = {s}"
    return text


def load_id_images(
    id_images_file_paths: list[Path], images_indices: list[tuple[int, int]]
) -> np.ndarray:
    """Loads the identification images from disk.

    Parameters
    ----------
    id_images_file_paths : list
        List of strings with the paths to the files where the images are
        stored.
    images_indices : list
        List of tuples (image_index, episode) that indicate each of the images
        to be loaded

    Returns
    -------
    Numpy array
        Numpy array of shape [number of images, width, height]
    """
    hdf5_datasets: list[h5py.Dataset] = []
    for path in id_images_file_paths:
        dataset = h5py.File(path, "r")["id_images"]
        assert isinstance(dataset, h5py.Dataset)
        hdf5_datasets.append(dataset)

    # Create entire output array
    test_image = hdf5_datasets[images_indices[0][1]][images_indices[0][0]]
    images = np.empty((len(images_indices), *test_image.shape), test_image.dtype)

    # Fill the output array
    for i, (image, episode) in enumerate(
        track(images_indices, "Loading identification images from the disk")
    ):
        images[i] = hdf5_datasets[episode][image]

    for hdf5_dataset in hdf5_datasets:
        hdf5_dataset.file.close()

    return images


def json_default(obj):
    """Encodes non JSON serializable object as dicts"""
    if isinstance(obj, Path):
        return {"py/object": "Path", "path": str(obj)}

    if isinstance(obj, (Timer, Episode)):
        dic = {"py/object": obj.__class__.__name__}
        dic.update(obj.__dict__)
        return dic

    if isinstance(obj, np.integer):
        return int(obj)

    if isinstance(obj, np.floating):
        return float(obj)

    if isinstance(obj, np.ndarray):
        return {"py/object": "np.ndarray", "values": obj.tolist()}

    if isinstance(obj, set):
        return {"py/object": "set", "values": list(obj)}

    raise ValueError(f"Could not JSON serialize {obj} of type {type(obj)}")


def json_object_hook(d: dict):
    """Decodes dicts from `json_default`"""
    if "py/object" in d:
        cls = d.pop("py/object")
        if cls == "Path":
            return Path(d["path"])
        if cls == "Episode":
            return Episode(**d)
        if cls == "Timer":
            return Timer.from_dict(d)
        if cls == "np.ndarray":
            return np.asarray(d["values"])
        if cls == "set":
            return set(d["values"])
        raise ValueError(f"Could not read {d}")
    return d


def resolve_path(path: Path | str) -> Path:
    return Path(path).expanduser().resolve()



from pathlib import Path

from idtrackerai import Video
from idtrackerai.utils import initLogger

from .general_video import generate_trajectories_video
from .individual_videos import generate_individual_video


def main():
    initLogger(check_version=False)
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "session_path",
        type=Path,
        help="Path to the video session created during the tracking session",
        metavar="",
    )

    parser.add_argument(
        "--individual",
        action="store_true",
        help="Generate individual video. Default is a general video",
    )
    parser.add_argument(
        "--gray", action="store_true", help="Draw the original video in grayscale"
    )

    parser.add_argument(
        "--t",
        type=Path,
        help=(
            "Path to the trajectory file, default is "
            "session_dir/trajectories/trajectories_wo_gaps.npy"
        ),
        default=None,
        metavar="",
    )
    parser.add_argument(
        "--tl",
        type=int,
        default=20,
        help=(
            "Trail length, number of points used to draw the individual trajectories"
            " traces"
        ),
        metavar="",
    )
    parser.add_argument(
        "--s", type=int, default=0, help="Frame where to start the video", metavar=""
    )
    parser.add_argument(
        "--e", type=int, default=None, help="Frame where to end the video", metavar=""
    )
    args = parser.parse_args()

    video = Video.load(args.session_path)

    if args.individual:
        generate_individual_video(
            video,
            args.t,
            draw_in_gray=args.gray,
            starting_frame=args.s,
            ending_frame=args.e,
        )
    else:
        generate_trajectories_video(
            video,
            args.t,
            draw_in_gray=args.gray,
            centroid_trace_length=args.tl,
            starting_frame=args.s,
            ending_frame=args.e,
        )


if __name__ == "__main__":
    main()



# This file is part of idtracker.ai a multiple animals tracking system
# described in [1].
# Copyright (C) 2017- Francisco Romero Ferrero, Mattia G. Bergomi,
# Francisco J.H. Heras, Robert Hinz, Gonzalo G. de Polavieja and the
# Champalimaud Foundation.
#
# idtracker.ai is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details. In addition, we require
# derivatives or applications to acknowledge the authors by citing [1].
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# For more information please send an email (idtrackerai@gmail.com) or
# use the tools available at https://gitlab.com/polavieja_lab/idtrackerai.git.
#
# [1] Romero-Ferrero, F., Bergomi, M.G., Hinz, R.C., Heras, F.J.H.,
# de Polavieja, G.G., Nature Methods, 2019.
# idtracker.ai: tracking all individuals in small or large collectives of
# unmarked animals.
# (F.R.-F. and M.G.B. contributed equally to this work.
# Correspondence should be addressed to G.G.d.P:
# gonzalo.polavieja@neuro.fchampalimaud.org)

import logging
import sys

import numpy as np
from rich.status import Status

from idtrackerai.utils import conf


class StopTraining:
    """Stops the training of the network according to the conditions specified
    in __call__
    """

    def __init__(
        self,
        epochs_before_checking_stopping_conditions=10,
        check_for_loss_plateau: bool = True,
        num_epochs=10,
    ):
        self.num_epochs = num_epochs  # maximal num of epochs
        self.number_of_classes = 2
        self.epochs_before_checking_stopping_conditions = (
            epochs_before_checking_stopping_conditions
        )
        self.overfitting_counter: int = 0
        """Number of epochs in which the network is overfitting before
        stopping the training"""

        self.check_for_loss_plateau = check_for_loss_plateau
        """if true the training is stopped if the loss is not decreasing enough"""

        self.epochs_completed = -1

    def __call__(
        self,
        loss_training: list,
        loss_validation: list,
        accuracy_validation: list,
        status: Status,
    ):
        self.epochs_completed += 1
        # check that the model did not diverged (nan loss).
        if self.epochs_completed > 0 and (
            np.isnan(loss_training[-1]) or np.isnan(loss_validation[-1])
        ):
            status.stop()
            logging.info(
                "The model diverged with loss NaN, falling back "
                "to detecting crossings with the model area"
            )
            return True
        # check if it did not reached the epochs limit
        if self.epochs_completed > self.num_epochs - 1:
            status.stop()
            logging.info(
                "The number of epochs completed is larger than the number "
                "of epochs set for training, we stop the training"
            )
            return True
        # check that the model is not overfitting or if it reached a
        # stable saddle (minimum)
        if self.epochs_completed > self.epochs_before_checking_stopping_conditions:
            current_loss = loss_validation[-1]
            previous_loss = np.nanmean(
                loss_validation[-self.epochs_before_checking_stopping_conditions : -1]
            )
            # The validation loss in the first 10 epochs could have
            # exploded but being decreasing.
            if np.isnan(previous_loss):
                previous_loss = sys.float_info[0]
            losses_difference = previous_loss - current_loss
            # check overfitting
            if losses_difference < 0.0:
                self.overfitting_counter += 1
                if self.overfitting_counter >= conf.OVERFITTING_COUNTER_THRESHOLD_DCD:
                    status.stop()
                    logging.info("Overfitting")
                    return True
            else:
                self.overfitting_counter = 0
            # check if the error is not decreasing much
            if self.check_for_loss_plateau and np.abs(
                losses_difference
            ) < conf.LEARNING_PERCENTAGE_DIFFERENCE_2_DCD * 10 ** (
                int(np.log10(current_loss)) - 1
            ):
                status.stop()
                logging.info(
                    "The losses difference is very small, we stop the training"
                )
                return True
            # if the individual accuracies in validation are 1. for all the animals
            if accuracy_validation[-1] == 1.0:
                status.stop()
                logging.info("The accuracy in validation is 100%, we stop the training")
                return True
            # if the validation loss is 0.
            if previous_loss == 0.0 or current_loss == 0.0:
                status.stop()
                logging.info("The validation loss is 0.0, we stop the training")
                return True

        return False



# This file is part of idtracker.ai a multiple animals tracking system
# described in [1].
# Copyright (C) 2017- Francisco Romero Ferrero, Mattia G. Bergomi,
# Francisco J.H. Heras, Robert Hinz, Gonzalo G. de Polavieja and the
# Champalimaud Foundation.
#
# idtracker.ai is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details. In addition, we require
# derivatives or applications to acknowledge the authors by citing [1].
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# For more information please send an email (idtrackerai@gmail.com) or
# use the tools available at https://gitlab.com/polavieja_lab/idtrackerai.git.
#
# [1] Romero-Ferrero, F., Bergomi, M.G., Hinz, R.C., Heras, F.J.H.,
# de Polavieja, G.G., Nature Methods, 2019.
# idtracker.ai: tracking all individuals in small or large collectives of
# unmarked animals.
# (F.R.-F. and M.G.B. contributed equally to this work.
# Correspondence should be addressed to G.G.d.P:
# gonzalo.polavieja@neuro.fchampalimaud.org)
from idtrackerai import ListOfBlobs, Video
from idtrackerai.utils import create_dir

from .crossing_detector import detect_crossings
from .model_area import ModelArea, compute_body_length


def crossings_detection_API(video: Video, list_of_blobs: ListOfBlobs) -> None:
    """
    This crossings detector works under the following assumptions
        1. The number of animals in the video is known (given by the user)
        2. There are frames in the video where all animals are separated from
        each other.
        3. All animals have a similar size
        4. The frame rate of the video is higher enough so that consecutive
        segmented blobs of pixels of the same animal overlap, i.e. some of the
        pixels representing the animal A in frame i are the same in the
        frame i+1.

    NOTE: This crossing detector sets the identification images that will be
    used to identify the animals
    """
    video.crossing_detector_timer.start()

    median_body_length = compute_body_length(
        list_of_blobs.blobs_in_video, video.number_of_animals
    )
    video.set_id_image_size(median_body_length)

    create_dir(video.id_images_folder, remove_existing=True)

    list_of_blobs.set_images_for_identification(
        video.episodes,
        video.id_images_file_paths,
        video.id_image_size,
        video.segmentation_data_folder,
    )
    list_of_blobs.compute_overlapping_between_subsequent_frames()
    create_dir(video.crossings_detector_folder)

    if not video.single_animal:
        model_area = ModelArea(list_of_blobs.blobs_in_video, video.number_of_animals)
        detect_crossings(list_of_blobs, video, model_area)
    else:
        for blob in list_of_blobs.all_blobs:
            blob.is_an_individual = True

    video.crossing_detector_timer.finish()



# This file is part of idtracker.ai a multiple animals tracking system
# described in [1].
# Copyright (C) 2017- Francisco Romero Ferrero, Mattia G. Bergomi,
# Francisco J.H. Heras, Robert Hinz, Gonzalo G. de Polavieja and the
# Champalimaud Foundation.
#
# idtracker.ai is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details. In addition, we require
# derivatives or applications to acknowledge the authors by citing [1].
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# For more information please send an email (idtrackerai@gmail.com) or
# use the tools available at https://gitlab.com/polavieja_lab/idtrackerai.git.
#
# [1] Romero-Ferrero, F., Bergomi, M.G., Hinz, R.C., Heras, F.J.H.,
# de Polavieja, G.G., Nature Methods, 2019.
# idtracker.ai: tracking all individuals in small or large collectives of
# unmarked animals.
# (F.R.-F. and M.G.B. contributed equally to this work.
# Correspondence should be addressed to G.G.d.P:
# gonzalo.polavieja@neuro.fchampalimaud.org)
import logging
from contextlib import suppress
from pathlib import Path

import numpy as np
from rich.console import Console
from torch.utils.data import DataLoader

from idtrackerai.network import LearnerClassification, NetworkParams, evaluate, train
from idtrackerai.tracker.accumulation_manager import AccumulationManager

from .stop_training_criteria import StopTraining


def TrainIdentification(
    learner: LearnerClassification,
    train_loader: DataLoader,
    val_loader: DataLoader,
    network_params: NetworkParams,
    stop_training: StopTraining,
    accumulation_manager: AccumulationManager | None = None,
) -> Path:
    logging.info("Training Identification Network")
    # TODO: Store accuracies and losses
    # store_training_accuracy_and_loss_data = \
    #     Store_Accuracy_and_Loss(
    #         self.network_params.save_folder,name='training')
    # store_validation_accuracy_and_loss_data = \
    #     Store_Accuracy_and_Loss(
    #         self.network_params.save_folder, ame='validation')

    # Initialize metric storage
    train_losses = []
    if network_params.loss in ("CEMCL", "CEMCL_weighted"):
        train_losses_CE = []
        train_losses_MCL = []
        val_losses_CE = []
        val_losses_MCL = []
    train_accs = []
    val_losses = []
    val_accs = []

    logging.debug("Entering the epochs loop...")
    with Console().status("[red]Epochs loop...") as status:
        while not stop_training(train_losses, val_losses, val_accs, status):
            epoch = stop_training.epochs_completed
            (loss, loss_CE, loss_MCL), train_acc = train(
                epoch, train_loader, learner, network_params
            )

            train_losses.append(loss)
            if network_params.loss in ("CEMCL", "CEMCL_weighted"):
                train_losses_CE.append(loss_CE)
                train_losses_MCL.append(loss_MCL)
            train_accs.append(train_acc)

            if val_loader is not None and (
                (not network_params.skip_eval) or (epoch == network_params.epochs - 1)
            ):
                loss, loss_CE, loss_MCL, val_acc = evaluate(
                    val_loader, None, network_params, learner
                )
                val_losses.append(loss)
                if network_params.loss in ("CEMCL", "CEMCL_weighted"):
                    val_losses_CE.append(loss_CE)
                    val_losses_MCL.append(loss_MCL)
                val_accs.append(val_acc)
            # Save checkpoint at each LR steps and the end of optimization
            # TODO: Consider saving only best model
            best_model_path = learner.snapshot(network_params.save_model_path)
            with suppress(IndexError):
                status.update(
                    f"[red]Epochs loop {epoch}: training loss = {train_losses[-1]:.6f},"
                    f" validation loss = {val_losses[-1]:.6f} and accuracy ="
                    f" {val_accs[-1]:.4%}"
                )

        logging.info("Last epoch loop: %s", status.status, extra={"markup": True})

    if np.isnan(train_losses[-1]) or np.isnan(val_losses[-1]):
        logging.warning(
            "The model diverged. Falling back to individual-crossing "
            "discrimination by average area model."
        )
    else:
        # update used_for_training flag to True for fragments used
        logging.info("Step completed.")
        if accumulation_manager is not None:
            accumulation_manager.update_fragments_used_for_training()

    return best_model_path



from pathlib import Path

import torch
from torch.nn import Module

from idtrackerai import Blob
from idtrackerai.crossings_detection.dataset.crossings_dataloader import (
    get_test_data_loader,
)


def get_predictions_crossigns(
    id_images_file_paths: list[Path], model: Module, blobs: list[Blob], use_gpu: bool
):
    loader = get_test_data_loader(id_images_file_paths, blobs)
    predictions = []

    model.eval()
    for input, _target in loader:
        # Prepare the inputs
        if use_gpu:
            with torch.no_grad():
                input = input.cuda()

        # Inference
        output = model(input)
        pred = output.argmax(1)  # find the predicted class

        predictions.extend(pred.cpu().numpy())

    return predictions



# This file is part of idtracker.ai a multiple animals tracking system
# described in [1].
# Copyright (C) 2017- Francisco Romero Ferrero, Mattia G. Bergomi,
# Francisco J.H. Heras, Robert Hinz, Gonzalo G. de Polavieja and the
# Champalimaud Foundation.
#
# idtracker.ai is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details. In addition, we require
# derivatives or applications to acknowledge the authors by citing [1].
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# For more information please send an email (idtrackerai@gmail.com) or
# use the tools available at https://gitlab.com/polavieja_lab/idtrackerai.git.
#
# [1] Romero-Ferrero, F., Bergomi, M.G., Hinz, R.C., Heras, F.J.H.,
# de Polavieja, G.G., Nature Methods, 2019.
# idtracker.ai: tracking all individuals in small or large collectives of
# unmarked animals.
# (F.R.-F. and M.G.B. contributed equally to this work.
# Correspondence should be addressed to G.G.d.P:
# gonzalo.polavieja@neuro.fchampalimaud.org)
import json
import logging
import sys
from copy import copy
from importlib import metadata
from math import sqrt
from pathlib import Path
from typing import Iterable

import cv2
import numpy as np

from .utils import (
    Episode,
    Timer,
    assert_all_files_exist,
    build_ROI_mask_from_list,
    check_if_identity_transfer_is_possible,
    conf,
    create_dir,
    json_default,
    json_object_hook,
    remove_dir,
    remove_file,
    resolve_path,
)


class Video:
    """
    A class containing the main features of the video.

    This class includes properties of the video by itself, user defined
    parameters for the tracking, and other properties that are generated
    throughout the tracking process.

    We use this class as a storage of data coming from different processes.
    However, this is bad practice and it will change in the future.
    """

    accumulation_step: int
    velocity_threshold: float
    erosion_kernel_size: int
    ratio_accumulated_images: float
    accumulation_folder: Path
    # FIXME it should depend on self.session_folder
    # return self.session_folder / f"accumulation_{self.accumulation_trial}"
    individual_fragments_stats: dict
    estimated_accuracy: float | None = None
    percentage_of_accumulated_images: list[float]
    # TODO: move to accumulation_manager.py
    accumulation_trial: int = 0
    # TODO: move to accumulation_manager.py
    session_folder: Path
    # TODO remove these defaults, they are already in __main__

    id_image_size: list[int]
    """ Shape of the Blob's identification images (width, height, n_channels)"""

    episodes: list[Episode]
    """Indicates the starting and ending frames of each video episode.
    Video episodes are used for parallelization of some processes"""

    video_paths: list[Path]
    """List of paths to the different files the video is composed of.
    If the video is a single file, the list will have length 1"""

    original_width: int
    """Original video width in pixels. It does not consider the resolution
    reduction factor defined by the user"""

    original_height: int
    """Original video width in pixels. It does not consider the resolution
    reduction factor defined by the user"""

    frames_per_second: int
    """Video frame rate in frames per second obtained by OpenCV from the
    video file"""

    number_of_error_frames: int = -1
    """The number of frames with more blobs than animals. Set on animals_detection."""

    accumulation_statistics: dict[str, list]

    accumulation_statistics_data: list[dict[str, list]]

    def __init__(
        self,
        video_paths: list[Path | str],
        number_of_animals,
        intensity_ths,
        area_ths,
        output_dir: Path | None,
        session,
        tracking_intervals: list | None,
        resolution_reduction: float,
        roi_list: list[str] | None,
        use_bkg: bool,
        track_wo_identities: bool,
        sigma_gaussian_blurring: float | None,
        check_segmentation: bool,
        identity_transfer: bool,
        knowledge_transfer_folder: Path | None,
        bkg_model,
        **kwargs,
    ):
        """Initializes a video object

        Parameters
        ----------
        video_path : str
            Path to a video file
        """
        if sigma_gaussian_blurring is None:
            sigma_gaussian_blurring = conf.SIGMA_GAUSSIAN_BLURRING
        if kwargs:
            logging.info(
                f"Ignoring the next arguments in Video.__init__():\n{kwargs.keys()}"
            )

        logging.debug("Initializing Video object")
        self.use_bkg = use_bkg
        self.check_segmentation = check_segmentation
        self.track_wo_identities = track_wo_identities
        """Flag indication the tracking will be performed without identities"""
        self.intensity_ths = intensity_ths
        self.area_ths = area_ths
        self.knowledge_transfer_folder = knowledge_transfer_folder
        self.resolution_reduction = resolution_reduction
        self.number_of_animals = int(number_of_animals)
        """Number of animals in the video indicated by user"""
        self.set_video_paths(video_paths)
        self.sigma_gaussian_blurring = sigma_gaussian_blurring
        self.data_policy: str = conf.DATA_POLICY
        self.frames_per_episode: int = conf.frames_per_episode
        self.version = metadata.version("idtrackerai")
        self.protocol3_action: str = conf.protocol3_action
        self.accumulation_statistics_data = [None] * (
            conf.MAXIMUM_NUMBER_OF_PARACHUTE_ACCUMULATIONS + 1
        )

        if self.knowledge_transfer_folder:
            self.knowledge_transfer_folder = Path(
                self.knowledge_transfer_folder
            ).resolve()
            assert (
                self.knowledge_transfer_folder.exists()
            ), f"{self.knowledge_transfer_folder} not found"

        self.original_width, self.original_height, self.frames_per_second = (
            self.get_info_from_video_paths(self.video_paths)
        )
        self.number_of_frames, _, self.tracking_intervals, self.episodes = (
            self.get_processing_episodes(
                self.video_paths, self.frames_per_episode, tracking_intervals
            )
        )

        logging.info(
            f"The video has {self.number_of_frames} "
            f"frames ({self.number_of_episodes} episodes)"
        )
        if len(self.episodes) < 10:
            for e in self.episodes:
                video_name = self.video_paths[e.video_path_index].name
                logging.info(
                    f"\tEpisode {e.index}, frames ({e.local_start} "
                    f"=> {e.local_end}) of /{video_name}"
                )
        assert self.number_of_episodes > 0

        if output_dir is not None:
            self.session_folder = (output_dir / f"session_{session.strip()}").resolve()
        else:
            self.session_folder = (
                self.video_folder / f"session_{session.strip()}"
            ).resolve()
        create_dir(self.session_folder)
        create_dir(self.preprocessing_folder)

        self.ROI_list = roi_list

        self.ROI_mask = build_ROI_mask_from_list(
            roi_list, self.original_width, self.original_height
        )

        if conf.IDENTIFICATION_IMAGE_SIZE > 0:
            self.id_image_size = [
                conf.IDENTIFICATION_IMAGE_SIZE,
                conf.IDENTIFICATION_IMAGE_SIZE,
                1,
            ]
        else:
            self.id_image_size = []

        if identity_transfer:
            # TODO: the id_image_size is not really passed by
            # the used but inferred from the knowledge transfer folder
            self.identity_transfer, self.id_image_size = (
                check_if_identity_transfer_is_possible(
                    self.number_of_animals, self.knowledge_transfer_folder
                )
            )
        else:
            self.identity_transfer = False

        self.bkg_model = bkg_model  # has a setter

        # Attributes computed by other processes in the tracking
        # During crossing detection
        self.median_body_length: float
        """median of the diagonals of individual blob's bounding boxes"""

        # TODO: move tracker.py
        self.first_frame_first_global_fragment = []  # updated later

        # During validation (in validation GUI)
        self.identities_groups = {}
        """Groups of identities stored during the validation of the tracking
        in the validation GUI. This is useful to group identities in different
        classes depending on the experiment.

        This feature was coded because some users require indicating classes
        of individuals but we do not use it in the lab."""

        self.setup_points: dict[str, list[tuple[float, float]]] = {}
        """Setup points"""

        self.identities_labels: list[str] = list(
            map(str, range(1, number_of_animals + 1))
        )
        """A list with a name for every identity. Defined ans used in validator"""

        # Flag to decide which type of interpolation is done. This flag
        # is updated when we update a blob centroid

        # Processes timers
        self.general_timer = Timer("Tracking session")
        self.detect_animals_timer = Timer("Animal detection")
        self.crossing_detector_timer = Timer("Crossing detection")
        self.fragmentation_timer = Timer("Fragmentation")
        self.tracking_timer = Timer("Tracking")
        self.protocol1_timer = Timer("Protocol 1")
        self.protocol2_timer = Timer("Protocol 2")
        self.protocol3_pretraining_timer = Timer("Protocol 3 pre-training")
        self.protocol3_accumulation_timer = Timer("Protocol 3 accumulation")
        self.identify_timer = Timer("Identification")
        self.impossible_jumps_timer = Timer("Impossible jumps correction")
        self.crossing_solver_timer = Timer("Crossings solver")
        self.create_trajectories_timer = Timer("Trajectories creation")

        self.general_timer.start()

    def __str__(self) -> str:
        return f"<session {self.session_folder}>"

    def set_id_image_size(self, median_body_length: int | float, reset=False):
        self.median_body_length = median_body_length
        if reset or not self.id_image_size:
            side_length = int(median_body_length / sqrt(2))
            side_length += side_length % 2
            self.id_image_size = [side_length, side_length, 1]
        logging.info(f"Identification image size set to {self.id_image_size}")

    @property
    def single_animal(self) -> bool:
        return self.number_of_animals == 1

    @property
    def bkg_model(self) -> np.ndarray | None:
        if self.background_path.is_file():
            return (
                cv2.imread(str(self.background_path))[..., 0].astype(np.float32)
                * self.bkg_norm
            )
        return None

    @bkg_model.setter
    def bkg_model(self, bkg: np.ndarray | None):
        if bkg is None:
            del self.bkg_model
        else:
            self.bkg_norm = bkg.max() / 255
            cv2.imwrite(
                str(self.background_path), (bkg / self.bkg_norm).astype(np.uint8)
            )
            logging.info(f"Background saved at {self.background_path}")

    @bkg_model.deleter
    def bkg_model(self):
        self.background_path.unlink(missing_ok=True)

    @property
    def ROI_mask(self) -> np.ndarray | None:
        if self.ROI_mask_path.is_file():
            return cv2.imread(str(self.ROI_mask_path))[..., 0].astype(bool)
        return None

    @ROI_mask.setter
    def ROI_mask(self, mask: np.ndarray | None):
        if mask is None:
            del self.ROI_mask
        else:
            cv2.imwrite(str(self.ROI_mask_path), (mask * 255).astype(np.uint8))
            logging.info(f"ROI mask saved at {self.ROI_mask_path}")

    @ROI_mask.deleter
    def ROI_mask(self):
        self.ROI_mask_path.unlink(missing_ok=True)

    def set_video_paths(self, video_paths: list[Path | str]):
        if not isinstance(video_paths, list):
            video_paths = [video_paths]
        self.assert_video_paths(video_paths)
        self.video_paths = [Path(path).expanduser().resolve() for path in video_paths]
        to_print = "Setting video_paths to:\n    " + "\n    ".join(
            map(str, self.video_paths)
        )
        logging.info(to_print)

    @property
    def video_folder(self) -> Path:
        """Directory where video was stored. Parent of video_path.

        Returns
        -------
        str
            Path to the video folder where the video to be tracked was stored.
        """
        return self.video_paths[0].parent

    @property
    def number_of_episodes(self):
        """Number of episodes in which the video is splitted for parallel
        processing.

        Returns
        -------
        int
            Number of parts in which the videos is splitted.

        See Also
        --------
        :int:`~idtrackerai.constants.FRAMES_PER_EPISODE`
        """
        return len(self.episodes)

    @property
    def width(self):
        """Video width in pixels after applying the resolution reduction
        factor.

        Returns
        -------
        int
            Video width in pixels after applying the resolution reduction
            factor defined by the user.
        """
        return np.round(self.original_width * self.resolution_reduction).astype(int)

    @property
    def height(self):
        """Video height in pixels after applying the resolution reduction
        factor.

        Returns
        -------
        int
            Video height in pixels after applying the resolution reduction
            factor.
        """
        return np.round(self.original_height * self.resolution_reduction).astype(int)

    # TODO: move to crossings_detection.py
    @property
    def median_body_length_full_resolution(self):
        """Median body length in pixels in full frame resolution
        (i.e. without considering the resolution reduction factor)
        """
        return self.median_body_length / self.resolution_reduction

    # Paths and folders
    # TODO: The different processes should create and store the path to the
    # folder where they save the data
    @property
    def preprocessing_folder(self) -> Path:
        return self.session_folder / "preprocessing"

    @property
    def background_path(self) -> Path:
        return self.preprocessing_folder / "background.png"

    @property
    def ROI_mask_path(self) -> Path:
        return self.preprocessing_folder / "ROI_mask.png"

    @property
    def trajectories_folder(self) -> Path:
        return self.session_folder / "trajectories"

    @property
    def crossings_detector_folder(self) -> Path:
        return self.session_folder / "crossings_detector"

    @property
    def pretraining_folder(self) -> Path:
        return self.session_folder / "pretraining"

    @property
    def individual_videos_folder(self) -> Path:
        return self.session_folder / "individual_videos"

    @property
    def auto_accumulation_folder(self) -> Path:
        return self.session_folder / f"accumulation_{self.accumulation_trial}"

    @property
    def id_images_folder(self) -> Path:
        return self.session_folder / "identification_images"

    # TODO: This should probably be the only path that should be stored in
    # Video.

    @property
    def blobs_path(self) -> Path:
        """get the path to save the blob collection after segmentation.
        It checks that the segmentation has been successfully performed"""
        return self.preprocessing_folder / "list_of_blobs.pickle"

    @property
    def blobs_no_gaps_path(self) -> Path:
        """get the path to save the blob collection after segmentation.
        It checks that the segmentation has been successfully performed"""
        return self.preprocessing_folder / "list_of_blobs_no_gaps.pickle"

    @property
    def blobs_path_validated(self) -> Path:
        return self.preprocessing_folder / "list_of_blobs_validated.pickle"

    @property
    def idmatcher_results_path(self) -> Path:
        return self.session_folder / "matching_results"

    @property
    def global_fragments_path(self) -> Path:
        """get the path to save the list of global fragments after
        fragmentation"""
        return self.preprocessing_folder / "list_of_global_fragments.pickle"

    @property
    def fragments_path(self) -> Path:
        """get the path to save the list of global fragments after
        fragmentation"""
        return self.preprocessing_folder / "list_of_fragments.pickle"

    @property
    def path_to_video_object(self) -> Path:
        return self.session_folder / "video_object.json"

    @property
    def ground_truth_path(self) -> Path:
        return self.video_folder / "_groundtruth.npy"

    @property
    def segmentation_data_folder(self) -> Path:
        return self.session_folder / "segmentation_data"

    @property
    def id_images_file_paths(self) -> list[Path]:
        return [
            self.id_images_folder / f"id_images_{e}.hdf5"
            for e in range(self.number_of_episodes)
        ]

    # Methods
    def save(self):
        """Saves the instantiated Video object"""
        logging.info(f"Saving video object in {self.path_to_video_object}")
        dict_to_save = copy(self.__dict__)
        dict_to_save.pop("episodes", None)
        dict_to_save.pop("_model_area", None)
        dict_to_save.pop("_accumulation_network_params", None)
        self.path_to_video_object.write_text(
            json.dumps(dict_to_save, default=json_default, indent=4)
        )

    @classmethod
    def load(cls, path: Path | str, video_paths_dir: Path | None = None) -> "Video":
        """Load a video object stored in a JSON file"""
        path = resolve_path(path)
        logging.info(f"Loading Video from {path}")
        if not path.is_file():
            path /= "video_object.json"
            if not path.is_file():
                path = path.with_suffix(".npy")
                if not path.is_file():
                    raise FileNotFoundError(f"{path} not found")

        if path.suffix == ".npy":
            video_dict = cls.open_from_v4(path)
        else:
            with open(path, "r", encoding="utf_8") as file:
                video_dict = json.load(file, object_hook=json_object_hook)

        video = cls.__new__(cls)
        video.__dict__.update(video_dict)
        video.update_paths(path.parent, video_paths_dir)
        try:
            (_, _, _, video.episodes) = video.get_processing_episodes(
                video.video_paths, video.frames_per_episode, video.tracking_intervals
            )
        except AttributeError:
            logging.warning(
                "Could not load video episodes probably due to loading an old version"
                " session"
            )
        return video

    @classmethod
    def open_from_v4(cls, path: Path) -> dict:
        from . import network

        logging.warning("Loading from v4: %s", path)

        # v4 compatibility
        sys.modules["idtrackerai.tracker.network.network_params"] = network
        _dict: dict = np.load(path, allow_pickle=True).item().__dict__
        del sys.modules["idtrackerai.tracker.network.network_params"]

        _dict["version"] = "4.0.12 or below"
        _dict["video_paths"] = list(map(Path, _dict.pop("_video_paths")))
        _dict["session_folder"] = path.parent
        _dict["median_body_length"] = _dict.pop("_median_body_length")
        _dict["frames_per_second"] = _dict.pop("_frames_per_second")
        _dict["original_width"] = _dict.pop("_original_width")
        _dict["original_height"] = _dict.pop("_original_height")
        _dict["number_of_frames"] = _dict.pop("_number_of_frames")
        _dict["identities_groups"] = _dict.pop("_identities_groups")
        _dict["id_image_size"] = list(_dict.pop("_identification_image_size"))
        _dict["setup_points"] = _dict.pop("_setup_points")
        _dict["number_of_animals"] = _dict["_user_defined_parameters"][
            "number_of_animals"
        ]
        _dict["tracking_intervals"] = _dict["_user_defined_parameters"][
            "tracking_interval"
        ]
        _dict["resolution_reduction"] = _dict["_user_defined_parameters"][
            "resolution_reduction"
        ]
        _dict["track_wo_identities"] = _dict["_user_defined_parameters"][
            "track_wo_identification"
        ]
        _dict["identities_labels"] = list(
            map(str, range(1, _dict["number_of_animals"] + 1))
        )
        _dict["accumulation_folder"] = (
            path.parent / Path(_dict.pop("_accumulation_folder")).name
        )
        _dict["_user_defined_parameters"].pop("mask")
        return _dict

    def update_paths(
        self, new_video_object_path: Path, user_video_paths_dir: Path | None = None
    ):
        """Update paths of objects (e.g. blobs_path, preprocessing_folder...)
        according to the new location of the new video object given
        by `new_video_object_path`.

        Parameters
        ----------
        new_video_object_path : str
            Path to a video_object.npy
        """
        logging.info(
            f"Searching video files: {[str(path.name) for path in self.video_paths]}"
        )
        folder_candidates: set[Path | None] = {
            user_video_paths_dir,
            self.video_paths[0],
            new_video_object_path,
            new_video_object_path.parent,
            self.session_folder.parent,
            self.session_folder,
        }

        for folder_candidate in folder_candidates:
            if folder_candidate is None:
                continue
            if folder_candidate.is_file():
                folder_candidate = folder_candidate.parent

            candidate_new_video_paths = [
                folder_candidate / path.name for path in self.video_paths
            ]

            try:
                assert_all_files_exist(candidate_new_video_paths)
            except FileNotFoundError:
                logging.info(f"Video files not found on {folder_candidate}")
            else:
                logging.info(f"All video files found on {folder_candidate}")
                found = True
                break
        else:
            found = False
            logging.error(f"Video file paths not found: {self.video_paths}")

        need_to_save = False
        if self.session_folder != new_video_object_path:
            logging.info(
                f"Updated session folder from {self.session_folder} to"
                f" {new_video_object_path}"
            )
            self.session_folder = new_video_object_path
            need_to_save = True

        if found and self.video_paths != candidate_new_video_paths:
            logging.info("Updating new video files paths")
            self.video_paths = candidate_new_video_paths
            need_to_save = True

        if need_to_save:
            self.save()

    @staticmethod
    def assert_video_paths(video_paths: Iterable[Path | str]):
        accepted_extensions = conf.AVAILABLE_VIDEO_EXTENSION
        assert video_paths, "Empty video_paths list"

        for path in video_paths:
            path = Path(path).expanduser().resolve()
            assert path.is_file(), f"Video file {path} not found"
            assert (
                path.suffix in accepted_extensions
            ), f"Supported video extensions are {accepted_extensions}"

    @staticmethod
    def get_info_from_video_paths(video_paths: Iterable[Path | str]):
        """Gets some information about the video from the video file itself.

        Returns:
            width: int, height: int, fps: int
        """

        widths, heights, fps = [], [], []
        for path in video_paths:
            cap = cv2.VideoCapture(str(path))
            widths.append(int(cap.get(3)))
            heights.append(int(cap.get(4)))

            try:
                fps.append(int(cap.get(5)))
            except cv2.error:
                logging.warning(f"Cannot read frame per second for {path}")
                fps.append(None)
            cap.release()

        assert len(set(widths)) == 1, "Video paths have different resolutions"
        assert len(set(heights)) == 1, "Video paths have different resolutions"
        if len(set(fps)) != 1:
            fps = [int(np.mean(fps))]
            logging.warning(
                f"Different frame rates detected ({fps}). "
                f"Setting the frame rate to the mean value: {fps[0]} fps"
            )

        return widths[0], heights[0], fps[0]

    # Methods to create folders where to store data
    # TODO: Some of these methods should go to the classes corresponding to
    # the process.

    def create_accumulation_folder(self, iteration_number=None, delete=False):
        """Folder in which the model generated while accumulating is stored
        (after pretraining)
        """
        if iteration_number is None:
            iteration_number = self.accumulation_trial
        self.accumulation_folder = (
            self.session_folder / f"accumulation_{iteration_number}"
        )
        # FIXME
        create_dir(self.accumulation_folder, remove_existing=delete)

    # Some methods related to the accumulation process
    # TODO: Move to accumulation_manager.py
    def init_accumulation_statistics_attributes(self):
        self.accumulation_statistics = {
            "n_accumulated_global_fragments": [],
            "n_non_certain_global_fragments": [],
            "n_randomly_assigned_global_fragments": [],
            "n_nonconsistent_global_fragments": [],
            "n_nonunique_global_fragments": [],
            "n_acceptable_global_fragments": [],
            "ratio_of_accumulated_images": [],
        }

    @staticmethod
    def get_processing_episodes(
        video_paths, frames_per_episode, tracking_intervals=None
    ) -> tuple[(int, list[int], list[list[int]], list[Episode])]:
        """Process the episodes by getting the number of frames in each video
        path and the tracking interval.

        Episodes are used to compute processes in parallel for different
        parts of the video. They are a tuple with
            (local start frame,
            local end frame,
            video path index,
            global start frame,
            global end frame)
        where "local" means in the specific video path and "global" means in
        the whole (multi path) video

        Episodes are guaranteed to belong to a single video path and to have
        all of their frames (end not included) inside a the tracking interval
        """

        def in_which_interval(frame_number, intervals) -> int | None:
            for i, (start, end) in enumerate(intervals):
                if start <= frame_number < end:
                    return i
            return None

        # total number of frames for every video path
        video_paths_n_frames = [
            int(cv2.VideoCapture(str(path)).get(7)) for path in video_paths
        ]
        number_of_frames = sum(video_paths_n_frames)

        # set full tracking interval if not defined
        if tracking_intervals is None:
            tracking_intervals = [[0, number_of_frames]]
        elif isinstance(tracking_intervals[0], int):
            tracking_intervals = [tracking_intervals]

        # find the global frames where the video path changes
        video_paths_changes = [0] + list(np.cumsum(video_paths_n_frames))

        # build an interval list like ("frame" refers to "global frame")
        #   [[first frame of video path 0, last frame of video path 0],
        #    [first frame of video path 1, last frame of video path 1],
        #    [...]]
        video_paths_intervals = list(
            zip(video_paths_changes[:-1], video_paths_changes[1:])
        )

        # find the frames where a tracking interval starts or ends
        tracking_intervals_changes = list(np.asarray(tracking_intervals).flatten())

        # Take into account tracking interval changes
        # and video path changes to compute episodes
        limits = video_paths_changes + tracking_intervals_changes

        # clean repeated limits and sort them
        limits = sorted(set(limits))

        # Create "long episodes" as the intervals between any video path
        # change or tracking interval change (keeping only the ones that
        # are inside a tracking interval)
        long_episodes = []
        for start, end in zip(limits[:-1], limits[1:]):
            if (
                in_which_interval(start, tracking_intervals) is not None
            ) and 0 <= start < number_of_frames:
                long_episodes.append((start, end))

        # build definitive episodes by dividing long episodes to fit in
        # the conf.FRAMES_PER_EPISODE restriction
        index = 0
        episodes = []
        for start, end in long_episodes:
            video_path_index = in_which_interval(start, video_paths_intervals)
            assert video_path_index is not None
            global_local_offset = video_paths_intervals[video_path_index][0]

            n_subepisodes = int((end - start) / (frames_per_episode + 1))
            new_episode_limits = np.linspace(start, end, n_subepisodes + 2, dtype=int)
            for new_start, new_end in zip(
                new_episode_limits[:-1], new_episode_limits[1:]
            ):
                episodes.append(
                    Episode(
                        index=index,
                        local_start=new_start - global_local_offset,
                        local_end=new_end - global_local_offset,
                        video_path_index=video_path_index,
                        global_start=new_start,
                        global_end=new_end,
                    )
                )
                index += 1
        return (number_of_frames, video_paths_n_frames, tracking_intervals, episodes)

    @staticmethod
    def in_which_interval(frame_number, intervals):
        for i, (start, end) in enumerate(intervals):
            if start <= frame_number < end:
                return i
        return None

    def delete_data(self):
        """Deletes some folders with data, to make the outcome lighter.

        Which folders are deleted depends on the constant DATA_POLICY
        """

        logging.info(f"Data policy: {self.data_policy}")

        if self.data_policy == "trajectories":
            remove_dir(self.segmentation_data_folder)
            remove_file(self.global_fragments_path)
            remove_dir(self.crossings_detector_folder)
            remove_dir(self.id_images_folder)
            for path in self.session_folder.glob("accumulation_*"):
                remove_dir(path)
            remove_dir(self.session_folder / "pretraining")
            remove_dir(self.preprocessing_folder)
        elif self.data_policy == "validation":
            remove_dir(self.segmentation_data_folder)
            remove_file(self.global_fragments_path)
            remove_dir(self.crossings_detector_folder)
            remove_dir(self.id_images_folder)
            for path in self.session_folder.glob("accumulation_*"):
                remove_dir(path)
            remove_dir(self.session_folder / "pretraining")
        elif self.data_policy == "knowledge_transfer":
            remove_dir(self.segmentation_data_folder)
            remove_file(self.global_fragments_path)
            remove_dir(self.crossings_detector_folder)
            remove_dir(self.id_images_folder)
        elif self.data_policy == "idmatcher.ai":
            remove_dir(self.segmentation_data_folder)
            remove_file(self.global_fragments_path)
            remove_dir(self.crossings_detector_folder)



# This file is part of idtracker.ai a multiple animals tracking system
# described in [1].
# Copyright (C) 2017- Francisco Romero Ferrero, Mattia G. Bergomi,
# Francisco J.H. Heras, Robert Hinz, Gonzalo G. de Polavieja and the
# Champalimaud Foundation.
#
# idtracker.ai is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details. In addition, we require
# derivatives or applications to acknowledge the authors by citing [1].
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# For more information please send an email (idtrackerai@gmail.com) or
# use the tools available at https://gitlab.com/polavieja_lab/idtrackerai.git.
#
# [1] Romero-Ferrero, F., Bergomi, M.G., Hinz, R.C., Heras, F.J.H.,
# de Polavieja, G.G., Nature Methods, 2019.
# idtracker.ai: tracking all individuals in small or large collectives of
# unmarked animals.
# (F.R.-F. and M.G.B. contributed equally to this work.
# Correspondence should be addressed to G.G.d.P:
# gonzalo.polavieja@neuro.fchampalimaud.org)
import json
import logging

import numpy as np
import torch
from torch.backends import cudnn

from idtrackerai import ListOfBlobs, ListOfFragments, ListOfGlobalFragments, Video
from idtrackerai.network import (
    LearnerClassification,
    NetworkParams,
    fc_weights_reinit,
    weights_xavier_init,
)
from idtrackerai.utils import CustomError, conf, create_dir, json_object_hook

from .accumulation_manager import AccumulationManager
from .accumulator import perform_one_accumulation_step
from .assigner import assign_remaining_fragments
from .identity_transfer import identify_first_global_fragment_for_accumulation
from .pre_trainer import pre_train_global_fragment


class TrackerAPI:
    def __init__(
        self,
        video: Video,
        list_of_blobs: ListOfBlobs,
        list_of_fragments: ListOfFragments,
        list_of_global_fragments: ListOfGlobalFragments,
    ):
        self.video = video
        self.list_of_blobs = list_of_blobs
        self.list_of_fragments = list_of_fragments
        self.list_of_global_fragments = list_of_global_fragments

        if self.video.knowledge_transfer_folder is not None:
            kt_info_dict_path = (
                self.video.knowledge_transfer_folder / "model_params.json"
            )
            try:
                self.knowledge_transfer_info_dict: dict = json.load(
                    kt_info_dict_path.open(), object_hook=json_object_hook
                )
            except FileNotFoundError:
                # Transferring from v4
                self.knowledge_transfer_info_dict: dict = np.load(
                    kt_info_dict_path.with_suffix(".npy"), allow_pickle=True
                ).item()
        else:
            self.knowledge_transfer_info_dict = {}

        # Old requirements for restoring
        self.processes_to_restore = {}

        self.accumulation_network_params: NetworkParams
        self.restoring_first_accumulation = False  # Flag restores first accumulation
        self.accumulation_step_finished = False  # Flag accumulation step finished

    def track_single_animal(self):
        logging.debug("Assigning identity 1 to all blobs")
        for bf in self.list_of_blobs.blobs_in_video:
            for blob in bf:
                blob.identity = 1
                blob.P2_vector = [1.0]

    def track_single_global_fragment_video(self):
        logging.info("TRACKING SINGLE GLOBAL FRAGMENT")

        def get_P2_vector(identity, number_of_animals):
            P2_vector = np.zeros(number_of_animals)
            P2_vector[identity - 1] = 1.0
            return P2_vector

        fragment_identifier_to_id = {}
        identity = 1
        for fragment in self.list_of_fragments.fragments:
            if fragment.is_an_individual:
                fragment_identifier_to_id[fragment.identifier] = identity
                identity += 1
            else:
                fragment_identifier_to_id[fragment.identifier] = None

        for bf in self.list_of_blobs.blobs_in_video:
            for b in bf:
                if b.is_an_individual:
                    b.identity = fragment_identifier_to_id[b.fragment_identifier]
                    b.P2_vector = get_P2_vector(
                        fragment_identifier_to_id[b.fragment_identifier],
                        self.video.number_of_animals,
                    )
        self.video.first_frame_first_global_fragment = [0]  # in case

    def track_with_identities(self) -> ListOfFragments:
        """In protocol 3, list_of_fragments is loaded from accumulation
        folders so the reference from outside tracker_API is lost.
        That's why list_of_fragments has to be returned"""
        self.video.tracking_timer.start()
        self._track_with_protocols_cascade()
        self.video.tracking_timer.finish()
        return self.list_of_fragments
        # track_with_cascade = True
        # if track_with_cascade:
        #     # This runs the protocol cascade and also the residual
        #     # identification, the impossible_jumps, the creation of
        #     # trajectories, the crossings interpolation, and the
        #     # creation of trajectories_wo_gaps
        #     # TODO: Factorize track_with_protocols_cascade so it only runs
        #     # up to residual identification
        #     self._track_with_protocols_cascade()
        # else:
        #     # TODO: Here is where new tracking methods should come
        #     # Call to tracking method

        #     # Call to postprocessing
        #     # TODO: Factorize postprocess_impossible_jumps
        #     # postprocess_impossible_jumps
        #     # create_trajectories
        #     # crossings_interpolation
        #     # create_trajectories_wo_gaps
        #     self.postprocess_impossible_jumps()
        #     raise NotImplementedError("New tracking methods are not allwoed")

    def _track_with_protocols_cascade(self):
        logging.info("******* Start tracking with protocol cascade ********")
        # Restoring

        delete = not self.processes_to_restore.get("protocols1_and_2")
        # Create accumulation folder
        self.video.create_accumulation_folder(iteration_number=0, delete=delete)

        self.init_accumulation_idCNN_params()

        # Restoring
        self.restoring_first_accumulation = False
        if self.processes_to_restore.get("post_processing"):
            raise NotImplementedError
            # self.restore_trajectories()
            # self.restore_crossings_solved()
            # self.restore_trajectories_wo_gaps()

        if self.processes_to_restore.get("residual_identification"):
            raise NotImplementedError
            # if self.video.track_wo_identities:
            # TODO: bring restoring back to life
            # raise
            # self.restore_trajectories()

            # else:
            # TODO: bring restoring back to life
            # raise
            # logging.info("Restoring residual identification")
            # self.restore_identification()
            # self.create_trajectories()

        if self.processes_to_restore.get("protocol3_accumulation"):
            raise NotImplementedError
            # logging.info("Restoring second accumulation")
            # # self.restore_second_accumulation()
            # self.video._first_frame_first_global_fragment = (
            #     self.video._first_frame_first_global_fragment
            # )
            # logging.warning(
            #     "first_frame_first_global_fragment "
            #     + str(
            #         self.video.first_frame_first_global_fragment
            #     )
            # )
            # logging.info("Starting identification")
            #
            # self.create_trajectories()

        if self.processes_to_restore.get("protocol3_pretraining"):
            # TODO: bring restoring back to life
            raise NotImplementedError
            # logging.info("Restoring pretraining")
            # logging.info("Initialising pretraining network")
            # self.init_pretraining_net()
            # logging.info("Restoring pretraining")
            # self.accumulation_step_finished = True
            # self.restore_first_accumulation()
            # self.restore_pretraining()
            # self.accumulation_manager.ratio_accumulated_images =
            # self.video.percentage_of_accumulated_images[0]
            # self.video._first_frame_first_global_fragment = [
            #     self.video._first_frame_first_global_fragment[
            #         0
            #     ]
            # ]
            # self.video._percentage_of_accumulated_images = [
            #     self.video.percentage_of_accumulated_images[0]
            # ]
            # logging.info("Start accumulation parachute")
            #
            # self.accumulate()

        if self.processes_to_restore.get("protocols1_and_2"):
            # TODO: bring restoring back to life
            raise NotImplementedError
            # logging.info("Restoring protocol 1 and 2")
            # self.restoring_first_accumulation = True
            # # self.restore_first_accumulation()
            # self.accumulation_manager.ratio_accumulated_images =
            # self.video.percentage_of_accumulated_images[0]
            # self.video._first_frame_first_global_fragment = [
            #     self.video._first_frame_first_global_fragment[
            #         0
            #     ]
            # ]
            # self.video._percentage_of_accumulated_images = [
            #     self.video.percentage_of_accumulated_images[0]
            # ]
            # self.accumulation_step_finished = True
            #
            # self.accumulate()

        if not self.processes_to_restore.get("protocols1_and_2"):
            logging.info("Starting protocol cascade")
            self.protocol1()

    def init_accumulation_idCNN_params(self):
        self.accumulation_network_params = NetworkParams(
            number_of_classes=self.video.number_of_animals,
            architecture=conf.IDCNN_NETWORK_NAME,
            save_folder=self.video.accumulation_folder,
            knowledge_transfer_folder=self.video.knowledge_transfer_folder,
            model_name="identification_network",
            image_size=self.video.id_image_size,
            scopes_layers_to_optimize=conf.LAYERS_TO_OPTIMISE_PRETRAINING,
            loss="CE",
            use_gpu=True,
            optimizer="SGD",
            schedule=[30, 60],
            optim_args={"lr": conf.LEARNING_RATE_IDCNN_ACCUMULATION, "momentum": 0.9},
            apply_mask=False,
            dataset="supervised",
            skip_eval=False,
            epochs=conf.MAXIMUM_NUMBER_OF_EPOCHS_IDCNN,
            return_store_objects=False,
        )
        # Save network params
        self.accumulation_network_params.save()

    def protocol1(self):
        self.video.protocol1_timer.start()

        # reset list of fragments and global fragments to fragmentation
        self.list_of_fragments.reset(roll_back_to="fragmentation")
        self.list_of_global_fragments.reset(roll_back_to="fragmentation")

        # Initialize idCNN
        logging.info("Setting learner class")
        self.learner_class = LearnerClassification
        logging.info("Creating idCNN")
        if self.video.knowledge_transfer_folder:
            try:
                self.identification_model = self.learner_class.load_model(
                    self.accumulation_network_params, scope="knowledge_transfer"
                )
                logging.info("Tracking with knowledge transfer")
                if not self.video.identity_transfer:
                    logging.info("Reinitializing fully connected layers")
                    self.identification_model.apply(fc_weights_reinit)
                else:
                    logging.info(
                        "Identity transfer. Not reinitializing the fully connected"
                        " layers."
                    )
            except RuntimeError:
                logging.error(
                    f"Could not load model {self.accumulation_network_params} to"
                    " transfer knowledge, following without knowledge nor identity"
                    " transfer"
                )
                self.learner_class = LearnerClassification
                self.identification_model = self.learner_class.create_model(
                    self.accumulation_network_params
                )
                self.identification_model.apply(weights_xavier_init)
        else:
            self.identification_model = self.learner_class.create_model(
                self.accumulation_network_params
            )
            self.identification_model.apply(weights_xavier_init)

        # Set first global fragment to start accumulation.
        # The network is passed in case of identity transfer.
        logging.info("Setting first global fragment for accumulation")
        first_global_fragment = (
            self.list_of_global_fragments.set_first_global_fragment_for_accumulation(
                accumulation_trial=0
            )
        )

        self.video.first_frame_first_global_fragment.append(
            first_global_fragment.first_frame_of_the_core
            if first_global_fragment is not None
            else None
        )
        if first_global_fragment is not None:
            identify_first_global_fragment_for_accumulation(
                first_global_fragment,
                self.video,
                network_params=self.accumulation_network_params,
                identification_model=self.identification_model,
                knowledge_transfer_info_dict=self.knowledge_transfer_info_dict,
            )

        # Order global fragments by distance to the first global fragment for the accumulation
        self.list_of_global_fragments.order_by_distance_to_the_first_global_fragment_for_accumulation(
            self.video.first_frame_first_global_fragment, accumulation_trial=0
        )

        # Instantiate accumulation manager
        self.accumulation_manager = AccumulationManager(
            self.video.id_images_file_paths,
            self.video.number_of_animals,
            self.list_of_fragments,
            self.list_of_global_fragments,
        )

        # Selecting the first global fragment is considered as
        # the 0 accumulation step
        self.accumulation_step_finished = True
        self.init_and_accumulate()

    def one_shot_accumulation(self):
        self.accumulation_step_finished = False
        assert self.identification_model is not None
        self.accumulation_manager.ratio_accumulated_images = (
            perform_one_accumulation_step(
                self.accumulation_manager,
                self.video,
                self.identification_model,
                self.learner_class,
                network_params=self.accumulation_network_params,
            )
        )
        self.accumulation_step_finished = True

    def accumulate(self):
        new_global_fragments_for_training = (
            self.accumulation_manager.new_global_fragments_for_training
        )
        if self.accumulation_step_finished and new_global_fragments_for_training:
            # Training and identification continues
            if (
                self.accumulation_manager.counter == 1
                and self.video.accumulation_trial == 0
            ):
                # first training finished
                self.video.protocol1_timer.finish()
                self.video.protocol2_timer.start()

            # Training and identification step
            self.one_shot_accumulation()
            # Re-enter the function for the next step of the accumulation
            self.accumulate()

        elif (
            not new_global_fragments_for_training
            and not self.video.protocol2_timer.finished
            and self.accumulation_manager.ratio_accumulated_images
            > conf.THRESHOLD_EARLY_STOP_ACCUMULATION
        ):
            # Accumulation stop because protocol 1 is successful
            self.save_after_first_accumulation()
            self.video.protocol1_timer.finish()
            logging.info("Protocol 1 successful")
            self.identify()

        elif (
            not new_global_fragments_for_training
            and not self.video.protocol3_pretraining_timer.finished
        ):
            logging.info("--------------------> No more new global fragments")
            self.save_after_first_accumulation()

            if (
                self.accumulation_manager.ratio_accumulated_images
                >= conf.THRESHOLD_ACCEPTABLE_ACCUMULATION
            ):
                self.video.protocol2_timer.finish()
                logging.info("Protocol 2 successful")
                self.identify()

            elif (
                self.accumulation_manager.ratio_accumulated_images
                < conf.THRESHOLD_ACCEPTABLE_ACCUMULATION
            ):
                self.video.protocol1_timer.finish()
                self.video.protocol2_timer.finish(raise_if_not_started=False)
                logging.warning(
                    "[red]Protocol 2 failed, protocol 3 is going to start",
                    extra={"markup": True},
                )
                ask_about_protocol3(
                    self.video.protocol3_action, self.video.number_of_error_frames
                )

                logging.warning("Going to start protocol 3")
                self.video.protocol3_pretraining_timer.start()

                self.pretraining_counter = 0
                self.protocol3()

        elif (
            self.video.protocol3_pretraining_timer.finished
            and self.video.accumulation_trial
            < conf.MAXIMUM_NUMBER_OF_PARACHUTE_ACCUMULATIONS
            and self.accumulation_manager.ratio_accumulated_images
            < conf.THRESHOLD_ACCEPTABLE_ACCUMULATION
        ):
            logging.info(
                "--------------------> Accumulation Protocol 3 failed. Opening"
                " parachute ..."
            )
            if self.video.accumulation_trial == 0:
                self.video.protocol3_accumulation_timer.start()
            self.video.accumulation_trial += 1
            if (
                not new_global_fragments_for_training
                and self.video.accumulation_trial > 1
            ):
                self.save_and_update_accumulation_parameters_in_parachute()
            self.accumulation_parachute_init(self.video.accumulation_trial)

            self.init_and_accumulate()

        elif self.video.protocol3_pretraining_timer.finished and (
            self.accumulation_manager.ratio_accumulated_images
            >= conf.THRESHOLD_ACCEPTABLE_ACCUMULATION
            or self.video.accumulation_trial
            >= conf.MAXIMUM_NUMBER_OF_PARACHUTE_ACCUMULATIONS
        ):
            logging.info("Accumulation after protocol 3 has been successful")
            self.video.protocol3_accumulation_timer.finish()

            self.save_after_second_accumulation()
            logging.info("Start residual identification")
            self.identify()

        # Whether to re-enter the function for the next accumulation step
        if self.accumulation_manager.new_global_fragments_for_training:
            self.accumulate()

    def init_and_accumulate(self):
        """
        This is called in the first step of each accumulation trial
        :param do_accumulate:
        :return:
        """
        logging.info("Entering accumulation loop")
        self.video.init_accumulation_statistics_attributes()
        self.accumulate()

    def save_after_first_accumulation(self):
        """Set flags and save data"""
        logging.info("Saving first accumulation parameters")

        if not self.restoring_first_accumulation:
            self.video.ratio_accumulated_images = (
                self.accumulation_manager.ratio_accumulated_images
            )
            self.video.percentage_of_accumulated_images = [
                self.video.ratio_accumulated_images
            ]
            self.video.save()
            self.list_of_fragments.save(self.video.fragments_path)
            self.list_of_global_fragments.save(self.video.global_fragments_path)

    """ pretraining """

    def protocol3(self):
        self.init_pretraining_variables()

        logging.info(
            "Starting pretraining. Checkpoints will be stored in %s"
            % self.video.pretraining_folder
        )

        if self.video.knowledge_transfer_folder:
            logging.info(
                "Performing knowledge transfer from %s"
                % self.video.knowledge_transfer_folder
            )
            self.pretrain_network_params.knowledge_transfer_folder = (
                self.video.knowledge_transfer_folder
            )

        logging.info("Start pretraining")
        self.pretraining_step_finished = True
        self.pretraining_loop()

    def init_pretraining_variables(self):
        self.init_pretraining_net()
        self.ratio_of_pretrained_images = 0

        # Initialize network
        self.learner_class = LearnerClassification
        if self.video.knowledge_transfer_folder:
            self.identification_model = self.learner_class.load_model(
                self.pretrain_network_params, scope="knowledge_transfer"
            )
            self.identification_model.apply(fc_weights_reinit)
        else:
            self.identification_model = self.learner_class.create_model(
                self.pretrain_network_params
            )
            self.identification_model.apply(weights_xavier_init)

    def init_pretraining_net(self):
        delete = not self.processes_to_restore.get("protocol3_pretraining")
        create_dir(self.video.pretraining_folder, remove_existing=delete)

        self.pretrain_network_params = NetworkParams(
            number_of_classes=self.video.number_of_animals,
            architecture=conf.IDCNN_NETWORK_NAME,
            save_folder=self.video.pretraining_folder,
            model_name="identification_network",
            image_size=self.video.id_image_size,
            scopes_layers_to_optimize=conf.LAYERS_TO_OPTIMISE_PRETRAINING,
            loss="CE",
            use_gpu=True,
            optimizer="SGD",
            schedule=[30, 60],
            optim_args={"lr": conf.LEARNING_RATE_IDCNN_ACCUMULATION, "momentum": 0.9},
            apply_mask=False,
            dataset="supervised",
            skip_eval=False,
            epochs=conf.MAXIMUM_NUMBER_OF_EPOCHS_IDCNN,
            return_store_objects=False,
        )

    def pretraining_loop(self):
        self.list_of_fragments.reset(roll_back_to="fragmentation")
        self.list_of_global_fragments.order_by_distance_travelled()
        self.one_shot_pretraining()
        self.continue_pretraining()

    def one_shot_pretraining(self):
        self.pretraining_step_finished = False
        self.pretraining_global_fragment = (
            self.list_of_global_fragments.global_fragments[self.pretraining_counter]
        )
        assert self.identification_model is not None
        (
            self.identification_model,
            self.ratio_of_pretrained_images,
            self.list_of_fragments,
            self.pretrained_model_path,
        ) = pre_train_global_fragment(
            self.video.number_of_animals,
            self.video.accumulation_step,
            self.identification_model,
            self.learner_class,
            self.pretrain_network_params,
            self.pretraining_global_fragment,
            self.list_of_fragments,
        )
        self.pretraining_counter += 1
        self.pretraining_step_finished = True

    def continue_pretraining(self, clock_unschedule=None):
        if (
            self.pretraining_step_finished
            and self.ratio_of_pretrained_images < conf.MAX_RATIO_OF_PRETRAINED_IMAGES
        ):
            self.one_shot_pretraining()

            if clock_unschedule is None:
                self.continue_pretraining()

        elif self.ratio_of_pretrained_images > conf.MAX_RATIO_OF_PRETRAINED_IMAGES:
            logging.warning("Calling accumulate from continue_pretraining")
            self.video.protocol3_pretraining_timer.finish()
            self.accumulate()

    """ parachute """

    def accumulation_parachute_init(self, iteration_number):
        logging.debug("------------------------> accumulation_parachute_init")
        logging.info("Starting accumulation %i" % iteration_number)

        delete = not self.processes_to_restore.get("protocol3_accumulation")

        self.video.create_accumulation_folder(
            iteration_number=iteration_number, delete=delete
        )
        self.video.accumulation_trial = iteration_number
        self.list_of_fragments.reset(roll_back_to="fragmentation")
        self.list_of_global_fragments.reset(roll_back_to="fragmentation")

        # Initialize network
        if self.video.identity_transfer:
            logging.info("Load model for identity transfer")
            self.identification_model = self.learner_class.load_model(
                self.accumulation_network_params
            )
        else:
            self.identification_model = None

        logging.info("Setting first global fragment for accumulation")
        first_global_fragment = (
            self.list_of_global_fragments.set_first_global_fragment_for_accumulation(
                accumulation_trial=iteration_number - 1
            )
        )

        self.video.first_frame_first_global_fragment.append(
            first_global_fragment.first_frame_of_the_core
            if first_global_fragment is not None
            else None
        )
        if first_global_fragment is not None:
            identify_first_global_fragment_for_accumulation(
                first_global_fragment,
                self.video,
                network_params=self.accumulation_network_params,
                identification_model=self.identification_model,
                knowledge_transfer_info_dict=self.knowledge_transfer_info_dict,
            )

        # Sort global fragments by distance
        self.list_of_global_fragments.order_by_distance_to_the_first_global_fragment_for_accumulation(
            self.video.first_frame_first_global_fragment,
            accumulation_trial=iteration_number - 1,
        )
        logging.warning(
            "first_frame_first_global_fragment "
            + str(self.video.first_frame_first_global_fragment)
        )
        logging.info(
            "We will restore the network from a previous pretraining: %s"
            % self.video.pretraining_folder
        )

        # Set saving folders
        self.accumulation_network_params.save_folder = self.video.accumulation_folder

        # Set restoring model_file
        self.accumulation_network_params.restore_folder = self.video.pretraining_folder

        # TODO: allow to train only the fully connected layers
        self.accumulation_network_params.scopes_layers_to_optimize = [
            "fully-connected1",
            "fully_connected_pre_softmax",
        ]
        logging.info("Initialising accumulation network")

        # Load pretrained network
        self.identification_model = self.learner_class.load_model(
            self.accumulation_network_params
        )

        # Re-initialize fully-connected layers
        self.identification_model.apply(fc_weights_reinit)

        # Instantiate accumualtion manager
        self.accumulation_manager = AccumulationManager(
            self.video.id_images_file_paths,
            self.video.number_of_animals,
            self.list_of_fragments,
            self.list_of_global_fragments,
        )

        logging.info("Start accumulation")

    def save_and_update_accumulation_parameters_in_parachute(self):
        logging.warning(
            "self.accumulation_manager.ratio_accumulated_images %.4f"
            % self.accumulation_manager.ratio_accumulated_images
        )
        self.video.ratio_accumulated_images = (
            self.accumulation_manager.ratio_accumulated_images
        )
        self.video.percentage_of_accumulated_images.append(
            self.video.ratio_accumulated_images
        )
        self.list_of_fragments.save(
            self.video.accumulation_folder / "list_of_fragments.pickle"
        )

    def save_after_second_accumulation(self):
        logging.info("Saving second accumulation parameters")
        # Save accumulation parameters
        self.save_and_update_accumulation_parameters_in_parachute()

        # Choose best accumulation
        self.video.accumulation_trial = int(
            np.argmax(self.video.percentage_of_accumulated_images)
        )

        # Update ratio of accumulated images and  accumulation folder
        self.video.ratio_accumulated_images = (
            self.video.percentage_of_accumulated_images[self.video.accumulation_trial]
        )
        self.video.create_accumulation_folder()

        # Load light list of fragments with identities of the best accumulation
        self.list_of_fragments = ListOfFragments.load(
            self.video.auto_accumulation_folder / "list_of_fragments.pickle"
        )

        # Save objects
        self.list_of_fragments.save(self.video.fragments_path)
        self.list_of_global_fragments.save(self.video.global_fragments_path)

        # set restoring folder
        logging.info("Restoring networks to best second accumulation")
        self.accumulation_network_params.restore_folder = self.video.accumulation_folder

        # TODO: allow to train only the fully connected layers
        self.accumulation_network_params.scopes_layers_to_optimize = [
            "fully-connected1",
            "fully_connected_pre_softmax",
        ]
        logging.info("Initialising accumulation network")

        # Load pretrained network
        self.identification_model = self.learner_class.load_model(
            self.accumulation_network_params
        )

        # # Re-initialize fully-connected layers
        # self.identification_model.apply(fc_weights_reinit)

        # Send model and criterion to GPU
        if self.accumulation_network_params.use_gpu:
            torch.cuda.set_device(0)
            logging.info(
                'Sending model and criterion to GPU: "%s"', torch.cuda.get_device_name()
            )
            cudnn.benchmark = True  # make it train faster
            self.identification_model = self.identification_model.cuda()

        self.video.save()

    """ Residual identification """

    def identify(self):
        self.video.identify_timer.start()
        assert self.identification_model is not None
        assign_remaining_fragments(
            self.list_of_fragments,
            self.identification_model,
            self.accumulation_network_params,
        )
        self.video.identify_timer.finish()


def ask_about_protocol3(protocol3_action: str, n_error_frames: int) -> None:
    """Raises a CustomError if protocol3_action is abort or aks and user answers abortion"""
    logging.info("Protocol 3 action: %s", protocol3_action)

    if protocol3_action == "abort":
        raise CustomError(
            "Protocol 3 was going to start but PROTOCOL3_ACTION is set to 'abort'"
        )
    if protocol3_action == "continue":
        return

    if protocol3_action != "ask":
        raise ValueError(
            f'PROTOCOL3_ACTION "{protocol3_action}" not in ("ask", "abort", "continue")'
        )

    if n_error_frames > 0:
        logging.info(
            (
                "Protocol 3 is a very time consuming algorithm and, in most cases, it"
                " can be avoided by redefining the segmentation parameters. As"
                " [red]there are %d frames with more blobs than animals[/red], we"
                " recommend you to abort the tracking session now and go back to the"
                " Segmentation app focusing on not having reflections, shades, etc."
                " detected as blobs. Check the following general recommendations:\n   "
                " - Define a region of interest to exclude undesired noise blobs\n    -"
                " Shrink the intensity (or background difference) thresholds\n    -"
                " Toggle the use of the background subtraction\n    - Shrink the blob's"
                " area thresholds"
            ),
            n_error_frames,
            extra={"markup": True},
        )
    else:
        logging.info(
            (
                "Protocol 3 is a very time consuming algorithm and, in most cases, it"
                " can be avoided by redefining the segmentation parameters. As"
                " [bold]there are NOT frames with more blobs than animals[/bold], the"
                " video is unlikely to have non-animal blobs. Even so, you can choose"
                " to abort the tracking session and redefine the segmentation"
                " parameters (specially shrinking the intensity (or background"
                " difference) thresholds) or to continue with Protocol 3."
            ),
            extra={"markup": True},
        )

    abort = None
    valid_answers = {"abort": True, "a": True, "continue": False, "c": False}
    while abort is None:
        answer_str = input(
            "What do you want to do now? Abort [A] or Continue [C]? "
        ).lower()
        if answer_str not in valid_answers:
            logging.warning("Invalid answer")
            continue
        abort = valid_answers[answer_str]
        logging.info("Answer --> Abort? %s", abort)
    if abort:
        raise CustomError(
            "This is not an actual error: protocol 3 was going to start"
            " but PROTOCOL3_ACTION is set to 'ask' and used aborted."
        )
    return



# This file is part of idtracker.ai a multiple animals tracking system
# described in [1].
# Copyright (C) 2017- Francisco Romero Ferrero, Mattia G. Bergomi,
# Francisco J.H. Heras, Robert Hinz, Gonzalo G. de Polavieja and the
# Champalimaud Foundation.
#
# idtracker.ai is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details. In addition, we require
# derivatives or applications to acknowledge the authors by citing [1].
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# For more information please send an email (idtrackerai@gmail.com) or
# use the tools available at https://gitlab.com/polavieja_lab/idtrackerai.git.
#
# [1] Romero-Ferrero, F., Bergomi, M.G., Hinz, R.C., Heras, F.J.H.,
# de Polavieja, G.G., Nature Methods, 2019.
# idtracker.ai: tracking all individuals in small or large collectives of
# unmarked animals.
# (F.R.-F. and M.G.B. contributed equally to this work.
# Correspondence should be addressed to G.G.d.P:
# gonzalo.polavieja@neuro.fchampalimaud.org)
import logging

import torch
from torch import nn
from torch.backends import cudnn
from torch.optim.lr_scheduler import MultiStepLR

from idtrackerai import GlobalFragment, ListOfFragments
from idtrackerai.network import LearnerClassification, NetworkParams, fc_weights_reinit
from idtrackerai.utils import conf

from .dataset.identification_dataloader import get_training_data_loaders
from .dataset.identification_dataset import split_data_train_and_validation
from .network.stop_training_criteria import StopTraining
from .network.trainer import TrainIdentification


def pre_train_global_fragment(
    number_of_animals: int,
    accumulation_step: int,
    identification_model: nn.Module,
    learner_class: type[LearnerClassification],
    network_params: NetworkParams,
    pretraining_global_fragment: GlobalFragment,
    list_of_fragments: ListOfFragments,
):
    """Performs pretraining on a single global fragments

    Parameters
    ----------
    net : <ConvNetwork obejct>
        an instance of the class :class:`~idCNN.ConvNetwork`
    pretraining_global_fragment : <GlobalFragment object>
        an instance of the class :class:`~globalfragment.GlobalFragment`
    list_of_fragments : <ListOfFragments object>
        an instance of the class :class:`~list_of_fragments.ListOfFragments`
    global_epoch : int
        global counter of the training epoch in pretraining
    check_for_loss_plateau : bool
        if True the stopping criteria (see :mod:`~stop_training_criteria`) will
        automatically stop the training in case the loss functin computed for
        the validation set of images reaches a plateau
    store_accuracy_and_error : bool
        if True the values of the loss function, accuracy and individual
        accuracy will be stored
    save_summaries : bool
        if True tensorflow summaries will be generated and stored to allow
        tensorboard visualisation of both loss and activity histograms
    store_training_accuracy_and_loss_data : <Store_Accuracy_and_Loss object>
        an instance of the class :class:`~Store_Accuracy_and_Loss`
    store_validation_accuracy_and_loss_data : <Store_Accuracy_and_Loss object>
        an instance of the class :class:`~Store_Accuracy_and_Loss`

    Returns
    -------
    <ConvNetwork object>
        network with updated parameters after training
    float
        ration of images used for pretraining over the total number of
        available images
    int
        global epoch counter updated after the training session
    <Store_Accuracy_and_Loss object>
        updated with the values collected on the training set of labelled
        images
    <Store_Accuracy_and_Loss object>
        updated with the values collected on the validation set of labelled
        images
    <ListOfFragments objects>
        list of instances of the class :class:`~fragment.Fragment`
    """
    # Get images and labels from the current global fragment
    images, labels = pretraining_global_fragment.get_images_and_labels(
        list_of_fragments.id_images_file_paths
    )

    train_data, val_data = split_data_train_and_validation(
        images, labels, validation_proportion=conf.VALIDATION_PROPORTION
    )
    logging.debug(f"images: {images.shape} {images.dtype}")
    logging.debug(f"labels: {labels.shape}")

    # Set data loaders
    train_loader, val_loader = get_training_data_loaders(
        number_of_animals, train_data, val_data
    )

    # Set criterion
    logging.info("Setting training criterion")
    criterion = nn.CrossEntropyLoss(weight=torch.tensor(train_data["weights"]))

    # Re-initialize fully-connected layers
    identification_model.apply(fc_weights_reinit)

    # Send model and criterion to GPU
    if network_params.use_gpu:
        torch.cuda.set_device(0)
        logging.info(
            'Sending model and criterion to GPU: "%s"', torch.cuda.get_device_name()
        )
        cudnn.benchmark = True  # make it train faster
        identification_model = identification_model.cuda()
        criterion = criterion.cuda()

    # Set optimizer
    logging.info("Setting optimizer")
    optimizer = torch.optim.__dict__[network_params.optimizer](
        identification_model.parameters(), **network_params.optim_args
    )

    # Set scheduler
    logging.info("Setting scheduler")
    scheduler = MultiStepLR(optimizer, milestones=network_params.schedule, gamma=0.1)

    # Set learner
    logging.info("Setting the learner")
    learner = learner_class(identification_model, criterion, optimizer, scheduler)

    # Set stopping criteria
    logging.info("Setting the stopping criteria")
    # set criteria to stop the training
    stop_training = StopTraining(
        network_params.number_of_classes,
        check_for_loss_plateau=True,
        first_accumulation_flag=accumulation_step == 0,
    )

    logging.info("Training identification network")
    best_model_path = TrainIdentification(
        learner, train_loader, val_loader, network_params, stop_training
    )

    logging.info("Identification network trained")

    for fragment in pretraining_global_fragment.individual_fragments:
        fragment.used_for_pretraining = True

    ratio_of_pretrained_images = (
        list_of_fragments.compute_ratio_of_images_used_for_pretraining()
    )
    logging.debug(
        "ratio of images used during pretraining: "
        f"{ratio_of_pretrained_images:.2%} (if higher than "
        f"{conf.MAX_RATIO_OF_PRETRAINED_IMAGES:.2%} we stop pretraining)"
    )

    return (
        identification_model,
        ratio_of_pretrained_images,
        list_of_fragments,
        best_model_path,
    )



# NetworkParams should be loaded before LearnerClassification
from idtrackerai.network.network_params import NetworkParams

from .evaluate import evaluate
from .learners import LearnerClassification
from .train import train
from .utils import Normalize, fc_weights_reinit, weights_xavier_init

__all__ = [
    "evaluate",
    "LearnerClassification",
    "train",
    "weights_xavier_init",
    "Normalize",
    "fc_weights_reinit",
    "NetworkParams",
]



# This file has been created to temporarily deal with coupling
# between accumulation_manager.py and list_of_global_fragments.py
# TODO remove this file and manage coupling properly
import numpy as np

from idtrackerai import Fragment, GlobalFragment


def get_P1_array_and_argsort(global_fragment: GlobalFragment):
    """Given a global fragment computes P1 for each of its individual
    fragments and returns a
    matrix of sorted indices according to P1

    Parameters
    ----------
    global_fragment : GlobalFragment object
        Collection of images relative to a part of the video in which all
        the animals are visible.

    Returns
    -------
    P1_array : nd.array
        P1 computed for every individual fragment in the global fragment
    index_individual_fragments_sorted_by_P1_max_to_min : nd.array
        Argsort of P1 array of each individual fragment
    """
    # get array of P1 values for the global fragment
    P1_array = np.asarray(
        [fragment.P1_vector for fragment in global_fragment.individual_fragments]
    )
    # get the maximum P1 of each individual fragment
    P1_max = np.max(P1_array, axis=1)
    # logging.debug("P1 max: %s" %str(P1_max))
    # get the index position of the individual fragments ordered by P1_max
    # from max to min
    index_individual_fragments_sorted_by_P1_max_to_min = np.argsort(P1_max)[::-1]
    return P1_array, index_individual_fragments_sorted_by_P1_max_to_min


def p1_below_random(
    P1_array: np.ndarray, index_individual_fragment: np.ndarray, fragment: Fragment
):
    """Evaluate if a fragment has been assigned with a certainty lower than
    random (wrt the number of possible identities)

    Parameters
    ----------
    P1_array  : nd.array
        P1 vector of a fragment object
    index_individual_fragment  : nd.array
        Argsort of the P1 array of fragment
    fragment : Fragment
        Fragment object containing images associated with a single individual

    Returns
    -------
    p1_below_random_flag : bool
        True if a fragment has been identified with a certainty below random
    """
    return (
        np.max(P1_array[index_individual_fragment, :]) < 1.0 / fragment.number_of_images
    )


def set_fragment_temporary_id(
    fragment: Fragment,
    temporary_id: int,
    P1_array: np.ndarray,
    index_individual_fragment: int,
):
    """Given a P1 array relative to a global fragment sets to 0 the row
    relative to fragment
    which is temporarily identified with identity temporary_id

    Parameters
    ----------
    fragment : Fragment
        Fragment object containing images associated with a single individual
    temporary_id : int
        temporary identifier associated to fragment
    P1_array  : nd.array
        P1 vector of fragment
    index_individual_fragment : int
        Index of fragment with respect to a global fragment in which it is
        contained

    Returns
    -------
    P1_array  : nd.array
        updated P1 array
    """
    fragment.temporary_id = int(temporary_id)
    P1_array[index_individual_fragment, :] = 0.0
    P1_array[:, temporary_id] = 0.0
    return P1_array



from typing import Optional

from PyQt6.QtCore import QEvent, QPointF, Qt
from PyQt6.QtGui import QKeyEvent, QPainterPath, QPalette, QResizeEvent
from PyQt6.QtWidgets import QFrame, QLabel, QSizePolicy, QWidget
from superqt import QLabeledRangeSlider, QLabeledSlider

from idtrackerai.utils import get_vertices_from_label


class QHLine(QFrame):
    def __init__(self):
        super().__init__()
        self.setFrameShape(QFrame.Shape.HLine)
        self.setContentsMargins(10, 0, 10, 0)
        self.setEnabled(False)

    def changeEvent(self, event: QEvent):
        if event.type() == QEvent.Type.EnabledChange:
            self.setEnabled(False)


def build_ROI_patches_from_list(
    width: int, height: int, list_of_ROIs: list[str] | None
) -> QPainterPath:
    path = QPainterPath()

    if list_of_ROIs is None:
        return path

    path.addRect(0, 0, width, height)

    for line in list_of_ROIs:
        points = get_vertices_from_label(line)
        path_i = QPainterPath(QPointF(*points[0]))
        for point in points[1:]:
            path_i.lineTo(*point)

        if line[0] == "+":
            path -= path_i.simplified()
        elif line[0] == "-":
            path += path_i.simplified()
        else:
            raise TypeError
    return path


class LabeledSlider(QLabeledSlider):
    def __init__(self, parent: QWidget, min, max):
        self.parent_widget = parent
        super().__init__(Qt.Orientation.Horizontal, parent)
        self.setRange(min, max)
        self.setSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Maximum)
        self._label.wheelEvent = lambda e: e.accept()

    def changeEvent(self, event: QEvent):
        super().changeEvent(event)
        if event.type() in (
            QEvent.Type.PaletteChange,
            QEvent.Type.EnabledChange,
            QEvent.Type.FontChange,
        ):
            style = (
                "QDoubleSpinBox{"
                + f"color: #{self.palette().color(QPalette.ColorRole.Text).rgba():x}"
                ";background:transparent; border: 0;"
                f" font-size:{self.font().pointSize()}pt"
                "}QDoubleSpinBox:!enabled{color: #"
                + f"{self.palette().color(QPalette.ColorGroup.Disabled, QPalette.ColorRole.Text).rgba():x}"
                ";}"
            )
            self._label.setStyleSheet(style)
            self._label._update_size()


class LabelRangeSlider(QLabeledRangeSlider):
    def __init__(
        self,
        min: int,
        max: int,
        parent: Optional[QWidget] = None,
        start_end_val: Optional[tuple[int, int]] = None,
        block_upper=True,
    ):
        self.parent_widget = parent
        super().__init__(Qt.Orientation.Horizontal, parent)
        self.setRange(min, max)
        self.setValue(start_end_val or (min, max))
        self.setSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Maximum)

        self._min_label.wheelEvent = lambda e: e.ignore()
        self._max_label.wheelEvent = lambda e: e.ignore()
        for handle in self._handle_labels:
            handle.wheelEvent = lambda event: event.ignore()

        self._min_label.setReadOnly(True)
        if block_upper:
            self._max_label.setReadOnly(True)

    def changeEvent(self, event: QEvent):
        super().changeEvent(event)
        if event.type() in (
            QEvent.Type.PaletteChange,
            QEvent.Type.EnabledChange,
            QEvent.Type.FontChange,
        ):
            style = (
                "QDoubleSpinBox{"
                + f"color: #{self.palette().color(QPalette.ColorRole.Text).rgba():x}"
                ";background:transparent; border: 0;"
                f" font-size:{self.font().pointSize()}pt"
                "}QDoubleSpinBox:!enabled{color: #"
                + f"{self.palette().color(QPalette.ColorGroup.Disabled, QPalette.ColorRole.Text).rgba():x}"
                ";}"
            )
            self._slider.setPalette(self.palette())
            self._min_label.setStyleSheet(style)
            self._max_label.setStyleSheet(style)
            self._max_label._update_size()
            self._min_label._update_size()
            for handle in self._handle_labels:
                handle.setStyleSheet(style)
                handle._update_size()

    def value(self) -> tuple[int, int]:
        return super().value()  # type: ignore


class WrappedLabel(QLabel):
    def __init__(
        self,
        text: str = "",
        framed: bool = False,
        align: Qt.AlignmentFlag = Qt.AlignmentFlag.AlignLeft,
    ):
        super().__init__(text)
        if framed:
            self.setBackgroundRole(QPalette.ColorRole.Base)
            self.setAutoFillBackground(True)
            self.setContentsMargins(5, 3, 5, 3)
        self.setAlignment(align)
        self.setWordWrap(True)
        self.setSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Maximum)

    def set_size(self):
        self.setMinimumHeight(0)
        self.setMinimumHeight(max(self.heightForWidth(self.width()), 1))

    def resizeEvent(self, a0: QResizeEvent):
        self.set_size()
        super().resizeEvent(a0)

    def setText(self, text: str):
        # Add Zero-width space in backslashes for proper word wrapping
        super().setText(text.replace("\\", "\\\u200b"))
        self.set_size()

    def text(self):
        return super().text().replace("\u200b", "")


def key_event_modifier(event: QKeyEvent) -> QKeyEvent | None:
    if event.key() == Qt.Key.Key_W:
        return QKeyEvent(event.type(), Qt.Key.Key_Up, event.modifiers())
    if event.key() == Qt.Key.Key_S:
        return QKeyEvent(event.type(), Qt.Key.Key_Down, event.modifiers())
    if event.key() in (Qt.Key.Key_D, Qt.Key.Key_A, Qt.Key.Key_Left, Qt.Key.Key_Right):
        # These keys would be accepted by QTableWidget
        # but we want them to control the VideoPlayer
        event.ignore()
        return None
    return event



import torch
from scipy.optimize import linear_sum_assignment as hungarian
from sklearn.metrics.cluster import (
    adjusted_mutual_info_score,
    adjusted_rand_score,
    normalized_mutual_info_score,
)
from torch import nn


def Class2Simi(x: torch.Tensor, mode="cls", mask=None):
    """
    Give a 1d torch tensor with classes in dense format, returns the pairwise similarity matrix liniarized. A mask can
    be applied to discard some elements of the similarity matrix.

    :param x: 1d torch tensor with classes in dense format
    :param mode: 'cls' for classification 'hinge' for clustering
    :param mask: 2d torch tensor with the mask to be applied to the pairwise similarity matrix
    :return: 1d torch tensor with the elements to be considered
    """
    # Convert class label to pairwise similarity
    n = x.nelement()
    assert (n - x.ndimension() + 1) == n, "Dimension of Label is not right"
    expand1 = x.view(-1, 1).expand(n, n)
    expand2 = x.view(1, -1).expand(n, n)
    out = expand1 - expand2
    out[out != 0] = -1  # dissimilar pair: label=-1
    out[out == 0] = 1  # Similar pair: label=1
    if mode == "cls":
        out[out == -1] = 0  # dissimilar pair: label=0
    if mode == "hinge":
        out = out.float()  # hingeloss require float type
    if mask is None:
        out = out.view(-1)
    else:
        mask = mask.detach()
        out = out[mask]
    return out


def weights_xavier_init(m):
    if isinstance(m, (nn.Linear, nn.Conv2d)):
        nn.init.xavier_uniform_(m.weight.data)


def fc_weights_reinit(m):
    if isinstance(m, nn.Linear):
        nn.init.xavier_uniform_(m.weight.data)


class Normalize:
    """Normalize a tensor image with mean and standard deviation.
    Given mean: ``(M1,...,Mn)`` and std: ``(S1,..,Sn)`` for ``n`` channels, this transform
    will normalize each channel of the input ``torch.*Tensor`` i.e.
    ``input[channel] = (input[channel] - mean[channel]) / std[channel]``
    .. note::
        This transform acts out of place, i.e., it does not mutates the input tensor.
    Args:
        mean (sequence): Sequence of means for each channel.
        std (sequence): Sequence of standard deviations for each channel.
        inplace(bool,optional): Bool to make this operation in-place.
    """

    # TODO: This is kind of a batch normalization but not trained. Explore using real BN in idCNN.

    def __init__(self, inplace=False):
        self.inplace = inplace  # TODO is inplace used?

    def __call__(self, tensor):
        """
        Args:
            tensor (Tensor): Tensor image of size (C, H, W) to be normalized.
        Returns:
            Tensor: Normalized Tensor image.
        """
        mean = torch.tensor([tensor.mean()])
        std = torch.tensor([tensor.std()])
        return tensor.sub_(mean[:, None, None]).div_(std[:, None, None])
        # return F.normalize(tensor, tensor.mean(), tensor.std(), self.inplace)


def prepare_task_target(target, args, mask=None):
    # Prepare the target for different criterion/tasks
    if args.loss == "CE":  # For standard classification
        if "semi" in args.dataset:
            one_hot_targets = target[:, :-1].reshape(-1)
            pairwise_targets = Class2Simi(target[:, -1], mode="hinge", mask=mask)
            train_target = torch.cat((one_hot_targets, pairwise_targets), 0)
            eval_target = target[:, -1]
        else:
            train_target = eval_target = target
    elif args.loss == "MCL":  # For clustering
        if "semi" in args.dataset:
            one_hot_targets = target[:, :-1].reshape(-1)
            pairwise_targets = Class2Simi(target[:, -1], mode="hinge", mask=mask)
            train_target = torch.cat((one_hot_targets, pairwise_targets), 0)
            eval_target = target[:, -1]
        else:
            train_target = Class2Simi(target, mode="hinge", mask=mask)
            eval_target = target
    elif args.loss in ("CEMCL", "CEMCL_weighted"):  # For semi-supervised clustering
        one_hot_targets = target[:, :-1].reshape(-1)
        pairwise_targets = Class2Simi(target[:, -1], mode="hinge", mask=mask)
        train_target = torch.cat((one_hot_targets, pairwise_targets), 0)
        eval_target = target[:, -1]
    return train_target, eval_target


class Confusion:
    """
    column of confusion matrix: predicted index
    row of confusion matrix: target index
    """

    def __init__(self, k, normalized=False):
        self.k = k
        self.conf = torch.LongTensor(k, k)
        self.normalized = normalized
        self.reset()

    def reset(self):
        self.conf.fill_(0)
        self.gt_n_cluster = None

    def cuda(self):
        self.conf = self.conf.cuda()

    def add(self, output, target):
        if target.size(0) > 1:
            output = output.squeeze_()
            target = target.squeeze_()
        assert output.size(0) == target.size(
            0
        ), "number of targets and outputs do not match"
        if output.ndimension() > 1:  # it is the raw probabilities over classes
            assert output.size(1) == self.conf.size(
                0
            ), "number of outputs does not match size of confusion matrix"

            _, pred = output.max(1)  # find the predicted class
        else:  # it is already the predicted class
            pred = output
        indices = (
            target * self.conf.stride(0) + pred.squeeze_().type_as(target)
        ).type_as(self.conf)
        ones = torch.ones(1).type_as(self.conf).expand(indices.size(0))
        self._conf_flat = self.conf.view(-1)
        self._conf_flat.index_add_(0, indices, ones)

    def classIoU(self, ignore_last=False):
        confusion_tensor = self.conf
        if ignore_last:
            confusion_tensor = self.conf.narrow(0, 0, self.k - 1).narrow(
                1, 0, self.k - 1
            )
        union = (
            confusion_tensor.sum(0).view(-1)
            + confusion_tensor.sum(1).view(-1)
            - confusion_tensor.diag().view(-1)
        )
        acc = confusion_tensor.diag().float().view(-1).div(union.float() + 1)
        return acc

    def recall(self, clsId):
        i = clsId
        TP = self.conf[i, i].sum().item()
        TPuFN = self.conf[i, :].sum().item()
        if TPuFN == 0:
            return 0
        return float(TP) / TPuFN

    def precision(self, clsId):
        i = clsId
        TP = self.conf[i, i].sum().item()
        TPuFP = self.conf[:, i].sum().item()
        if TPuFP == 0:
            return 0
        return float(TP) / TPuFP

    def f1score(self, clsId):
        r = self.recall(clsId)
        p = self.precision(clsId)
        if (p + r) == 0:
            return 0
        return 2 * float(p * r) / (p + r)

    def acc(self):
        TP = self.conf.diag().sum().item()
        total = self.conf.sum().item()
        if total == 0:
            return 0.0
        return float(TP) / total

    def optimal_assignment(self, gt_n_cluster=None, assign=None):
        if assign is None:
            mat = -self.conf.cpu().numpy()  # hungaian finds the minimum cost
            r, assign = hungarian(mat)
        self.conf = self.conf[:, assign]
        self.gt_n_cluster = gt_n_cluster
        return assign

    def show(self, width=6, row_labels=None, column_labels=None):
        print("Confusion Matrix:")
        conf = self.conf
        rows = self.gt_n_cluster or conf.size(0)
        cols = conf.size(1)
        if column_labels is not None:
            print(("%" + str(width) + "s") % "", end="")
            for c in column_labels:
                print(("%" + str(width) + "s") % c, end="")
            print("")
        for i in range(0, rows):
            if row_labels is not None:
                print(("%" + str(width) + "s|") % row_labels[i], end="")
            for j in range(0, cols):
                print(("%" + str(width) + ".d") % conf[i, j], end="")
            print("")

    def conf2label(self):
        conf = self.conf
        gt_classes_count = conf.sum(1).squeeze()
        n_sample = gt_classes_count.sum().item()
        gt_label = torch.zeros(n_sample)
        pred_label = torch.zeros(n_sample)
        cur_idx = 0
        for c in range(conf.size(0)):
            if gt_classes_count[c] > 0:
                gt_label[cur_idx : cur_idx + gt_classes_count[c]].fill_(c)
            for p in range(conf.size(1)):
                if conf[c][p] > 0:
                    pred_label[cur_idx : cur_idx + conf[c][p]].fill_(p)
                cur_idx = cur_idx + conf[c][p]
        return gt_label, pred_label

    def clusterscores(self):
        target, pred = self.conf2label()
        NMI = normalized_mutual_info_score(target, pred, average_method="arithmetic")
        ARI = adjusted_rand_score(target, pred)
        AMI = adjusted_mutual_info_score(target, pred, average_method="arithmetic")
        return {"NMI": NMI, "ARI": ARI, "AMI": AMI}
